
import io
import matplotlib.pyplot as plt


from .preprocessing_for_piechart import preprocessing_for_piechart
from .preprocessing_for_phases import preprocessing_for_phases
from .plot_adaptive_ring_with_phase import plot_adaptive_ring_with_phase


def generate_piechart_with_phase(df_meta_with_value,
                      colors=None,
                      hatches=None,
                      reading_interval_in_mins=None,
                      asset_group=None,
                      minimum_cnt_for_protection=None,
                      max_display_cnt=None,
                      sum_column=None, 
                      pct_column=None,
                      gene_column=None,
                      fontsize=None,
                      directory_to_savefig=None
                      ):

    df_sort_idx_trim = preprocessing_for_piechart(df_meta_with_value,
                                                reading_interval_in_mins=reading_interval_in_mins,
                                                asset_group=asset_group,
                                                minimum_cnt_for_protection=minimum_cnt_for_protection,
                                                    max_display_cnt=max_display_cnt,
                                                    sum_column=sum_column, 
                                                    pct_column=pct_column,
                                                    gene_column=gene_column)
    
    phase_values_list = preprocessing_for_phases(df_meta_with_value,
                                                 reading_interval_in_mins=reading_interval_in_mins)

    plot_adaptive_ring_with_phase(df_sort_idx_trim, phase_values_list,
                                    colors=colors,
                                    hatches=hatches,
                                    sum_column=sum_column, 
                                    pct_column=pct_column,
                                    gene_column=gene_column,
                                    fontsize=fontsize,
                                    )

    png_name = 'consumption_by_assetclass_piechart_with_phase.png'
    if directory_to_savefig == None:
        png_object = io.BytesIO()
        plt.savefig(png_object, format='png', dpi=200)
        plt.close()
        return {png_name: png_object}
    else:
        png_path = directory_to_savefig + png_name
        plt.savefig(png_path, format='png', dpi=200)
        plt.close()
        return {png_name: png_path}




def generate_day_code_list(datetime_list):
    
    previous_day_month = None
    previous_day_year = None
    day_code_list = []

    len_dates = len(datetime_list)

    for item in datetime_list:
        weekday_str = item.day_name()[0]

        # handle the case that for monthly reports, the day numbers will be very close to each other
        if (len_dates < 20) or (weekday_str=='M'): 
            day_str = "\n"+str(item.day)
        else:
            day_str = "\n "

        # the month and year are only added when it starts or changes

        if item.month == previous_day_month:
            month_str = ""
        else:
            month_str = " "+str(item.strftime("%b"))
        previous_day_month = item.month

        if item.year == previous_day_year:
            year_str = ''
        else:
            year_str = " "+str(item.year)
        previous_day_year = item.year

        day_code = weekday_str+day_str+'\n'+month_str+year_str

        day_code_list.append(day_code)
        
    return day_code_list



from .MondayProcessorException import MondayProcessorException
from .MondayProcessorBoardNotFound import MondayProcessorBoardNotFound
from .MondayProcessorNoColumInBoard import MondayProcessorNoColumInBoard
from .MondayProcessorNoGroupsInBoard import MondayProcessorNoGroupsInBoard
from .MondayProcessorNoItemInGroup import MondayProcessorNoItemInGroup
from .MondayProcessorBoardColumnNotMatch import MondayProcessorBoardColumnNotMatch
from .MondayProcessorPageNotSupported import MondayProcessorPageNotSupported





def get_entries_to_display(df_pivot_asset_group_by_period_renamed_rdx,
                           max_display_cnt=None,
                           gene_column=None,
                           asset_group=None):

    if max_display_cnt is None:
        max_display_cnt = 8
    
    if gene_column is None:
        gene_column = 'generator'

    if asset_group is None:
        asset_group = 'thing_category'

    # get the rows to keep/display and redundant rows to merge
    df_pivot_asset_group_by_period_renamed_rdx['to_be_merged'] = df_pivot_asset_group_by_period_renamed_rdx['rank_overall_protect'] > max_display_cnt
    df_pivot_asset_group_by_period_renamed_rdx_to_merge = df_pivot_asset_group_by_period_renamed_rdx.loc[df_pivot_asset_group_by_period_renamed_rdx['to_be_merged']]
    df_pivot_asset_group_by_period_renamed_rdx_to_keep = df_pivot_asset_group_by_period_renamed_rdx.loc[~df_pivot_asset_group_by_period_renamed_rdx['to_be_merged']]
    df_pivot_asset_group_by_period_renamed_rdx_to_merge_grp = df_pivot_asset_group_by_period_renamed_rdx_to_merge.groupby([gene_column]).agg({'sum': 'sum', 
                                                                                                                                              'sum_pre': 'sum', 
                                                                                                                                              'sub': 'sum'})

    # merge the redundant rows to "others" in the rows to keep
    for gene, sr_gene in df_pivot_asset_group_by_period_renamed_rdx_to_merge_grp.iterrows():

        idx_last_row = df_pivot_asset_group_by_period_renamed_rdx_to_keep.loc[df_pivot_asset_group_by_period_renamed_rdx_to_keep[gene_column]==gene].index.max()
        df_pivot_asset_group_by_period_renamed_rdx_to_keep.loc[idx_last_row, asset_group] = 'Other Assets' if gene==False else 'Other Sources'

        for column in df_pivot_asset_group_by_period_renamed_rdx_to_merge_grp.columns:
            df_pivot_asset_group_by_period_renamed_rdx_to_keep.loc[idx_last_row, column] += sr_gene[column]

    return df_pivot_asset_group_by_period_renamed_rdx_to_keep



def get_pos_bot_and_ylim(df_pivot_working_hours_sorted_reset,
                         pad_pct_above_max = None):

    if pad_pct_above_max is None:
        pad_pct_above_max = 0.2

    # make the index unique for later processing
    df_pivot_working_hours_sorted_reset_pos = df_pivot_working_hours_sorted_reset[False]
    sr_pivot_working_hours_sorted_reset_bot = df_pivot_working_hours_sorted_reset[True].sum(axis=1)

    # get the maximum and minimum values
    zero_line = 0
    tick_value_e_max = max(zero_line, df_pivot_working_hours_sorted_reset.sum(axis=1).max()) # ensure that the 0 line is still showing
    tick_value_e_min = min(zero_line, sr_pivot_working_hours_sorted_reset_bot.min()) # ensure that the 0 line is still showing
    tick_value_e_var = tick_value_e_max - tick_value_e_min

    # add padding to the boundary values to make the range
    
    if abs(tick_value_e_min) > (tick_value_e_max*0.01): # if the abs value of tick_value_e_min is very small, then set zero as the lowest range
        pad_pct_below_min = 0.05
    else:
        pad_pct_below_min = 0

    tick_range_e_max = tick_value_e_max + tick_value_e_var*pad_pct_above_max
    tick_range_e_min = tick_value_e_min - tick_value_e_var*pad_pct_below_min

    ls_ylim = [tick_range_e_min, tick_range_e_max]

    return df_pivot_working_hours_sorted_reset_pos, sr_pivot_working_hours_sorted_reset_bot, ls_ylim




def statement_for_biggest_ooh(df_asset_group_period_sum_others, number_for_pick_out=3,  period_freq_str='week'):

    # print('df_asset_group_period_sum_others: ', df_asset_group_period_sum_others) # debug

    if (number_for_pick_out+1) > df_asset_group_period_sum_others.shape[0]: # there are two rows for "Total" and "Others", this is to avoid them to be shown.

        number_for_pick_out = df_asset_group_period_sum_others.shape[0] - 2

        print(f'[WARN]: in func [statement_for_biggest_ooh], the no. of entries in df_asset_group_period_sum_others [{df_asset_group_period_sum_others.shape[0]}] is smaller then the expected number_for_pick_out [ {number_for_pick_out}]', number_for_pick_out)

    df_ooh_biggest = df_asset_group_period_sum_others.head(number_for_pick_out+1).tail(number_for_pick_out).drop(columns=['gt_pct_others','sum_for_sort'])
    
    # pdb.set_trace()

    statement = f"""The largest out-of-hours consumers of energy over the previous {period_freq_str} were: """ #temp fix

    for index, item in enumerate(df_ooh_biggest['sum'].round().astype('int').items()):

        if item[0] not in ['Total', 'Others']: # only add valid entries, not total and others.

            statement_item = "  \t \t \t \t"+str(index+1)+'. '+item[0]+' '+str(item[1])+' kwh,'
            
            statement += statement_item
        else:
            print(f'[WARN]: statement_for_biggest_ooh - Please note that ooh category [{item[0]}] has been skipped')

    statement = statement[:-1]+'.'
    
    return statement


import pandas as pd

def load_org(engine, table_name):

    with engine.connect() as conn:

        query = f""" select id as org_id, name as org from {table_name} """ # meta.space_annotations

        df_org = pd.read_sql_query(query, con=conn) 

        df_org.org = df_org.org.str.strip()
        
    return df_org


def check_table_existence(engine, table_name):
    
    table_schema_name_list = table_name.split('.')

    if len(table_schema_name_list) > 1:
        table_schema = table_schema_name_list[0]
        table_non_schema_name = table_schema_name_list[1]
    else:
        table_schema = 'public'
        table_non_schema_name = table_name

    query_check_existing_tables = f"""select table_name from information_schema.tables where table_schema = '{table_schema}' 
    and table_name = '{table_non_schema_name}'"""

    with engine.connect() as conn:
        list_tables = conn.execute(query_check_existing_tables)

    table_name_list = [table_name[0] for table_name in list_tables]

    if len(table_name_list) == 1:
        # print(f'The table [{table_name}] already exists')
        return True
    else:
        # print(f'The table [{table_name}] does not exist')
        return False





def get_formatted_pct(pct_value, 
                      pct_len=None):
    
    if pct_len is None:
        pct_len = 2

    # format the relative change

    pad_for_3_digit = 1

    if abs(pct_value) <= 1.06:
        label_pct_str = str(round(abs(pct_value)*100))
        label_pct_unit = r"%"
        label_pct_dot = " "*label_pct_str.count('.')
        label_pct_pad = ' ' *int((pct_len - len(label_pct_str))*2 + pad_for_3_digit) 
    elif abs(pct_value) < 1000:
        label_pct_str = str(round(abs(pct_value))) # todo: config number of decimals
        if len(label_pct_str) < 2:
            label_pct_str = str(round(abs(pct_value)*10)/10) # add one decimal
        label_pct_unit = r"X"
        label_pct_dot = " "*label_pct_str.count('.')
        label_pct_pad = ' ' *int((pct_len - len(label_pct_str))*2 +1 + pad_for_3_digit)
    elif abs(pct_value) < 1000_000:
        label_pct_str = str(round(abs(pct_value)/1000)) # todo: config number of decimals
        if len(label_pct_str) < 2:
            label_pct_str = str(round(abs(pct_value)/1000*10)/10) # add one decimal
        label_pct_unit = r"KX"
        label_pct_dot = " "*label_pct_str.count('.')
        label_pct_pad = ' ' *int((pct_len - len(label_pct_str))*2 -1 + pad_for_3_digit)
    else:
        label_pct_str = str(round(abs(pct_value)/1000_000)) # todo: config number of decimals
        if len(label_pct_str) < 2:
            label_pct_str = str(round(abs(pct_value)/1000_1000*10)/10) # add one decimal
        label_pct_unit = r"MX"
        label_pct_dot = " "*label_pct_str.count('.')
        label_pct_pad = ' ' *int((pct_len - len(label_pct_str))*2 -2 + pad_for_3_digit)

    # print(f'label_pct_pad = "{label_pct_pad}"')
    label_pct = label_pct_dot + label_pct_pad + label_pct_str + label_pct_unit + ',' + ' ' * 5

    return label_pct



def get_start_time(self):

        period_range = self.get_period_range()
        start_time = period_range[-(self.period_offset_previous+1)].start_time # start time of the (previous) period to be compared
        return start_time


from datetime import timedelta
import pandas as pd
# from sqlalchemy import create_engine

from mdl.utils_ import enriching_time_features

def statement_for_avg_action_time(db, building_name, thing_name, start_time, end_time,
                                  action=1,
                                  period_freq_str='week'):

    

    time_restriction = f"""(time >= '{start_time}') and (time < '{end_time}')"""

    statement_list = [f""""root_name"='{building_name}'"""]
    statement_full = " and ".join(statement_list)

    query = f"""select time,action,root_name,thing_name from {db.table_on_off.name} where {statement_full} and {time_restriction};"""

    engine = db.engine

    with engine.connect() as conn:
        df_on_off = pd.read_sql_query(query, con=conn)

    df_on_off['thing_name'] = df_on_off['thing_name'].str.lstrip().str.rstrip("0123456789 ")

    df_on_off_selected = df_on_off.loc[df_on_off['thing_name']==thing_name]
    

    if df_on_off_selected.shape[0] > 0: # handle the  case that no on/off data is returned

        df_on_off_selected = enriching_time_features(df_on_off_selected)

        df_on_off_avg = df_on_off_selected.groupby(['action']).time_of_day_in_float.mean()

        avg_start_time = str(timedelta(hours=df_on_off_avg[action])).split('.')[0][:-3]

        # AM/PM to the time

        time_list = avg_start_time.split(":")

        hour_digit = int(time_list[0])
        if hour_digit < 12:
            letter = " AM" 
        else: 
            letter = " PM"
            if hour_digit >= 13:
                time_list[0] = str(hour_digit-12)

        time_list.append(letter)
        time_list[1] = ":"+time_list[1]
        avg_start_time_with_letter = "".join(time_list)   
        
        start_finish_dict = {1: 'start', -1: 'finish'}


        statement = f"{avg_start_time_with_letter} was the average {start_finish_dict[action]} time for the {thing_name} over this {period_freq_str}." # temp fix
    else: 
        statement = None
    
    return statement




import inspect
from airflow.hooks.base_hook import BaseHook
import json

def get_conn_by_id_airflow(conn_id):

    cred_conn = BaseHook.get_connection(conn_id)

    cred_dict = {}
    for attribute in dir(cred_conn):
        print(cred_conn, 'attr: ', attribute)

        if not attribute.startswith('__'):
            if not inspect.ismethod(attribute):
                cred_dict[attribute] = getattr(cred_conn, attribute)
                if attribute == 'schema':
                    cred_dict['database'] = getattr(cred_conn, attribute)
                elif attribute == 'login':
                    cred_dict['username'] = getattr(cred_conn, attribute)
                    cred_dict['client_id'] = getattr(cred_conn, attribute)
        
    cred_dict['password'] = cred_conn.get_password()
    cred_dict['token'] = cred_conn.get_password()
    cred_dict['api_key'] = cred_conn.get_password()
    cred_dict['secret'] = cred_conn.get_password()

    cred_extra = cred_conn.get_extra()

    if len(cred_extra) > 2:
        cred_extra_json = json.loads(cred_extra)

        cred_dict = {**cred_dict, **cred_extra_json}

    print('cred_dict: ', cred_dict)
    
    return cred_dict





    # server = email_conn.host
    # port = email_conn.port
    # send_from = json.loads(email_conn.get_extra())["send_from"]
    # username = email_conn.login
    # password = email_conn.get_password()


import pathlib

def generate_dag_id(file):

    current_path = pathlib.Path(file)
    current_dir = current_path.parents[1]
    current_dir_str = str(current_dir)

    # directory = current_dir_str.split('_dags')[-1][1:].replace('/', '_')

    project = current_dir_str.split('_pkg')[0].split('/')[-1]

    dag_id = project

    # if directory == '':
    #     dag_id = project
    # else:
    #     dag_id = f"{project}_{directory}"

    return dag_id


from typing import List

from .MondayProcessorExceptions_ import (
    MondayProcessorBoardNotFound,
    MondayProcessorNoColumInBoard,
    MondayProcessorBoardColumnNotMatch
)

def _searchColumnIdsByName(self, board_id:int, column_name:str) -> List[str]:
    data = self.monday.boards.fetch_columns_by_board_id(board_ids=board_id)
    boards = data['data']['boards']
    if(len(boards) <1):
        raise MondayProcessorBoardNotFound(f"Monday board id ({board_id}) not found when perform MondayProcessor._searchConfColumnIds")
    columns = boards[0]['columns']
    if(len(columns) < 1):
        raise MondayProcessorNoColumInBoard(f"No column in board {board_id}")
    #  = boards[0][groups]
    # print("columns: ", columns)
    ColumnId = self._searchConfColumnIdsMatch(columns=columns,includeList=[column_name])
    if (ColumnId == None): raise MondayProcessorBoardColumnNotMatch("cannot find a column with this name")

    ret = [ColumnId]
    return ret




from .get_legend_handles import get_legend_handles
from .generate_legend_labels import generate_legend_labels

def show_legend(ax, df_pre_pie, 
                fontsize=None,
                sum_column=None, 
                pct_column=None):
    
    if fontsize is None:
        fontsize = 13

    valid_handles = get_legend_handles(ax)
    legend_labels = generate_legend_labels(df_pre_pie, 
                                            sum_column=sum_column, 
                                            pct_column=pct_column)

    x0, y0, width, height = 1.15, 0.5, 0, 0 
    leg_ncol = 1
    leg_loc = 'center left'

    leg = ax.legend(handles=valid_handles,
                    labels=legend_labels,
                    loc=leg_loc,
                    facecolor='w', 
                    edgecolor='w',
                    borderaxespad=0,
                    bbox_to_anchor=(x0, y0, width, height), 
                    fontsize=fontsize,  
                    ncol=leg_ncol, 
                    handleheight=1.2, 
                    labelspacing=0.6, 
                    title=None)

    leg.set_in_layout(False)



import os
import pandas as pd

pd.options.mode.chained_assignment = None

from .prepare_report import prepare_report
from .email_report import email_report
from .ClientSite import ClientSite
from .load_config import load_config
from .select_sites_and_emails import select_sites_and_emails

from mdl.utils_ import (
    delete_files_in_directory,
    setup_table
)

def prepare_and_email_reports(cf, 
                              mode='review', 
                              debug=False,
                              sync_source=None,
                              cache_in_memory=False, 
                              cache_directory=None):

    if cache_in_memory:
        report_file_folder = None
        directory_to_savefile = None
    else:
        if cache_directory == None:
            cache_directory = os.path.join(os.path.dirname(__file__), '_cache_/')

        report_file_folder = os.path.join(cache_directory, '_cached_reports_/')
        directory_to_savefile = os.path.join(cache_directory, '_cached_files_/')

    setup_table(cf.postgresdb.engine, 
                cf.postgresdb.table_mailing.name, 
                cf.postgresdb.table_mailing.columns_dict,
                hypertable=cf.postgresdb.table_mailing.hypertable)
    
    setup_table(cf.postgresdb.engine, 
                cf.postgresdb.table_cache.name, 
                cf.postgresdb.table_cache.columns_dict,
                hypertable=cf.postgresdb.table_cache.hypertable)
    
    # get contacts from db

    df_contact = load_config(cf.postgresdb.engine, 
                             cf.sharepoint,
                             cf.postgresdb.table_org.name,
                             cf.postgresdb.table_building.name,
                             cf.postgresdb.table_timezone.name,
                             cf.postgresdb.table_sqm.name,
                             cf.postgresdb.table_price.name,
                             cf.postgresdb.table_cache.name, 
                             list(cf.postgresdb.table_cache.columns_dict),
                             sync_source=sync_source)

    df_contact_select = select_sites_and_emails(df_contact, mode, cf.email.reviewer_emails)

    # print(f'{df_contact_select=}')

    # quit()
    for _, row in df_contact_select.iterrows():

        site_obj = ClientSite(row)
        df_receivers = pd.DataFrame({'email': row['send_to']})

        if df_receivers['email'].nunique() > 0:
            
            file_path_or_obj_dict, current_period_obj = prepare_report(cf, site_obj, 
                                                                    debug=debug,
                                                                    report_file_folder=report_file_folder, 
                                                                    directory_to_savefile=directory_to_savefile)    

            # print('row: ', row)
            # print("row['send_to']: ", row['send_to'])

            df_receivers = pd.DataFrame({'email': row['send_to']})
            df_receivers['name'] = site_obj.manager_name

            receivers = df_receivers.to_dict('records')

            # print('receivers: ', receivers)

            # quit()

            email_report(receivers, cf.smtp, cf.email, current_period_obj, file_path_or_obj_dict)
        else: 
            print(f'{site_obj.site_name} - {site_obj.site_name}: no email address to send to!')
        
    if not cache_in_memory:
        delete_files_in_directory(report_file_folder, ignore=['__init__.py'])
        delete_files_in_directory(directory_to_savefile, ignore=['__init__.py'])



from .column_values import column_values

def items(column_values_ids):
    items = f"""
                items{{
                        id
                        name
                        {column_values(column_values_ids)}
                        }}
            """
    return items


import pandas as pd

def enriching_time_features(df_meta_with_value, period_freq='M', 
                            schedule_records = None,
                            ooh_values=['OUT OF HOURS', 'POST TRADE', 'closed'],
                            weekend=None, 
                            working_end_time=None, 
                            working_start_time=None,
                            timezone='UTC'):
    
    if weekend == None:
        weekend = 5
    if working_end_time == None:
        working_end_time = "18:00:00"
    if working_start_time == None:
        working_start_time = "08:00:00"

    # manipulate and clean the data

    # print('df_meta_with_value.head(3).time: ', df_meta_with_value.head(3).time)

    df_meta_with_value.time = pd.to_datetime(df_meta_with_value.time).dt.tz_convert(timezone)

    # print('df_meta_with_value.head(3): ', df_meta_with_value.head(3)) # test
    df_meta_with_value = df_meta_with_value.set_index("time") 

    # enrich_time_information
    df_meta_with_value["date"] = df_meta_with_value.index.date
    df_meta_with_value["day_of_month"] = df_meta_with_value.index.day
    df_meta_with_value["time_of_day"] = df_meta_with_value.index.time

    df_meta_with_value['time_of_day_in_float'] = df_meta_with_value.index.hour+df_meta_with_value.index.minute/60+df_meta_with_value.index.second/3600

    df_meta_with_value["weekday"] = df_meta_with_value.index.weekday
    df_meta_with_value["day_name"] = df_meta_with_value.index.day_name()
    df_meta_with_value["day_code"] = df_meta_with_value["day_name"].str[0]
    df_meta_with_value["month"] = df_meta_with_value.index.month
    df_meta_with_value["month_name"] = df_meta_with_value.index.month_name() #new change/implementation -RP

    # print('schedule_records: ', schedule_records)

    if schedule_records is not None:

        df_meta_with_value["out_of_hours"] = False

        for schedule_record in schedule_records:
            # print('schedule_record: ', schedule_record)
            if schedule_record['name'] in ooh_values:
                df_meta_with_value["out_of_hours"] |= (df_meta_with_value['day_name'] == schedule_record['day']) & \
                                                (df_meta_with_value["time_of_day"] >= schedule_record['start_time']) & \
                                                (df_meta_with_value["time_of_day"] < schedule_record['end_time'])
                
                # print('ooh exist:', sr_ooh_record.unique())
                ## For out of hour - POST TRADE, OUT OF HOUR
                ## For In Hour - PRE TRADE, TRADE

                # If out of hour - then run the below files,  checkpoint - POST TRADE & OUT OF HOUR
                # if !out of hour - checkpoint - POST TRADE | OUT OF HOUR, run different file
    else:
        df_meta_with_value["out_of_hours"] = df_meta_with_value['weekday'].ge(weekend) | \
                                                (df_meta_with_value["time_of_day"] > pd.to_datetime(working_end_time).time()) | \
                                                (df_meta_with_value["time_of_day"] < pd.to_datetime(working_start_time).time())
    
    df_meta_with_value["period"] = df_meta_with_value.index.tz_localize(None).to_period(freq=period_freq)
    
    return df_meta_with_value




def get_start_time_co2_barchart(self):

    period_range = self.get_period_range()
    start_time_co2_barchart = period_range.min().start_time # start time of the first bar in the co2 barchart
    return start_time_co2_barchart




def get_formatted_sum(value, 
                     sum_value_len=None):
    
    if sum_value_len is None:
        sum_value_len = 8

    # adjust the decimal point and the unit based on the actual value
    sum_abs = abs(value)
    if sum_abs >= 99_950: # KWH
        label_kwh_str = f'{value/1000:4.0f} MWh'
    elif sum_abs >= 1000: # KWH
        label_kwh_str = f'{value/1000:4.1f} MWh'
    elif sum_abs >= 99.95: # KWH 
        label_kwh_str = f'{value:4.0f} KWh' # to handle the case that (sum_abs:  99.98276) => (100.0 KWh)
    elif sum_abs >= 1:
        label_kwh_str = f'{value:4.1f} KWh'
    elif sum_abs >= 0.09995:
        label_kwh_str = f'{value*1000:4.0f} Wh'
    else:
        label_kwh_str = f'{value*1000:4.1f} Wh'

    label_kwh_str_strip = label_kwh_str.strip()

    label_kwh_pad_digi = ' ' *int((sum_value_len - len(label_kwh_str_strip))*2)
    label_kwh_pad_dot = ' '*label_kwh_str_strip.count('.')

    label_kwh = label_kwh_pad_dot + label_kwh_pad_digi + label_kwh_str_strip + ','

    return label_kwh



from office365.sharepoint.client_context import ClientContext
from office365.runtime.auth.client_credential import ClientCredential

from .get_excel_file_from_link import get_excel_file_from_link
from .load_excel_as_df import load_excel_as_df


def load_mailing_list(sharepoint):

    # get sharepoint context
    ctx = ClientContext(sharepoint.site_url).with_credentials(ClientCredential(sharepoint.client_id, sharepoint.client_secret))

    file_name = getattr(sharepoint.excel, 'file_name', None)
    sheet_name = getattr(sharepoint.excel, 'sheet', None)

    if file_name is None:
        raise Exception('[ERROR]: Please specify the excel file name!')

    sharepoint_link = sharepoint.folder_url + '/' + file_name 

    excel_file = get_excel_file_from_link(ctx, sharepoint_link)

    # sheet_name = sharepoint.excel.sheet
    
    df_contact = load_excel_as_df(ctx, excel_file, 
                                  sheet_name=sheet_name)

    return df_contact






def get_end_time(self):

    period_range = self.get_period_range()
    # end_time = period_range.max().start_time # end time of the most recent finished period
    end_time = period_range[-(self.period_offset)].start_time # end time of the offset period
    return end_time


from .items_page import items_page

def groups(id, column_values_ids):
    groups = f'''
            groups(ids:"{id}"){{
                            id
                            {items_page(column_values_ids)}

                        }}
            '''
    return groups



import uuid

def uuid_for_nona(value):
    
    if isinstance(value, str):
        if value == ' ':
            return None
        else:
            return uuid.UUID(value)
    else:
        return None


import pandas as pd

from MondayProcessor import MondayProcessor
from cache_return import cache_return

@cache_return
def load_mailing_list(monday_token, board_id, 
                      columns_concerned=None, 
                      email_prefix=None):
    
    if columns_concerned is None:
        from .columns_concerned import columns_concerned


    email_prefix = email_prefix or 'email'

    mp = MondayProcessor(monday_token)

    groups_dict = mp._fetchGroups(board_id)

    columns_data = mp.monday.boards.fetch_columns_by_board_id(board_ids=board_id)

    columns_all = [column['title'] for column in columns_data['data']['boards'][0]['columns']]
    len_ep = len(email_prefix)
    columns_emails = [column for column in columns_all if column[:len_ep]==(email_prefix[:len_ep])]

    missing_columns = []
    for concerned_column in columns_concerned:
        if concerned_column not in columns_all:
            missing_columns.append(concerned_column)

    if len(missing_columns) > 0:
        raise Exception(f'[ERROR]: There are missing columns: {missing_columns}!')
    
    columns_select = columns_concerned + columns_emails # merge the two lists
    

    # print('columns_all: ', columns_all)
    # print('columns_select: ', columns_select)

    df_board = pd.DataFrame([])
    for group_id in groups_dict:
        
        board_results = mp._fetchColumnValueByGroup(board_id=board_id, groupId=group_id, columns=columns_select)

        df_group = pd.DataFrame(board_results)
        df_group['group_name'] = groups_dict[group_id]
        df_board = pd.concat([df_board, df_group])

    return df_board



import matplotlib.pyplot as plt


def set_x_label_prepare_xtick(ax, df_grouped_working_hours_period_unstacked, 
                              fontsize=None):
    
    if fontsize is None:
        fontsize = 13

    if df_grouped_working_hours_period_unstacked.index.freqstr == 'M':
        x_ticks_labels = df_grouped_working_hours_period_unstacked.index.strftime("%b %y").tolist()
        ax.set_xlabel("Month", labelpad = 13, fontsize = fontsize)
        plt.xticks(rotation=45) # please note: plt cannot be replace by ax, as it will throw errors
    else:
        ls_week_str = (df_grouped_working_hours_period_unstacked.index).strftime("%W").tolist()
        x_ticks_labels = [f"{(int(week)+1):02d}" for week in ls_week_str]   # add 1 week to the current_period_obj so that the start week is not Week 00
        ax.set_xlabel("Week Number", labelpad = 13, fontsize = fontsize)
        plt.xticks(rotation=0)

    x_ticks_labels.insert(0,"")
    x_ticks_labels.append("")

    return x_ticks_labels




# ruler annotation shinking config

config_dict = {'full_name': {'size_lim': 0.11*2, 
                             'names': {'part_one': "", 
                                       'part_two': '', 
                                       'part_three': ''}},
               'short_name':  {'size_lim': 0.035*2, 
                               'names': {'part_one': "", 
                                       'part_two': '', 
                                       'part_three': ''}},
               'null_name':  {'size_lim': 0.03*2, 
                              'names': {'part_one': "", 
                                        'part_two': '', 
                                        'part_three': ''}},
               }


def ruler_anno_shink_cf(bar_width, part):

    if bar_width > config_dict['full_name']['size_lim']:
        rpa_text = config_dict['full_name']['names'][part]
    elif bar_width > config_dict['short_name']['size_lim']:
        rpa_text = config_dict['short_name']['names'][part]
    else:
        rpa_text = config_dict['null_name']['names'][part]

    return rpa_text



from typing import List

from .MondayProcessorExceptions_ import (
    MondayProcessorBoardNotFound,
    MondayProcessorNoColumInBoard,
    MondayProcessorBoardColumnNotMatch
)

def _searchConfColumnIds(self, board_id:int) -> List[str]:
    data = self.monday.boards.fetch_columns_by_board_id(board_ids=board_id)
    boards = data['data']['boards']
    if(len(boards) <1):
        raise MondayProcessorBoardNotFound(f"Monday board id ({board_id}) not found when perform MondayProcessor._searchConfColumnIds")
    columns = boards[0]['columns']
    if(len(columns) < 1):
        raise MondayProcessorNoColumInBoard(f"No column in board {board_id}")
    #  = boards[0][groups]
    ctColumnId = self._searchConfColumnIdsMatch(columns=columns,includeList=["ct","type"])
    if (ctColumnId == None): raise MondayProcessorBoardColumnNotMatch("cannot find CT type column")
    phaseColumnId = self._searchConfColumnIdsMatch(columns=columns,includeList=["L1","L2","L3"], excludeList=["supply"])
    if (phaseColumnId == None): raise MondayProcessorBoardColumnNotMatch("cannot find circuit phase column")
    channelColumnId = self._searchConfColumnIdsMatch(columns=columns,includeList=["channel","no"])
    if (channelColumnId == None): raise MondayProcessorBoardColumnNotMatch("cannot find PRISM channel column")
    snColumnId = self._searchConfColumnIdsMatch(columns=columns,includeList=["ID","PRISM"])
    if (snColumnId == None): raise MondayProcessorBoardColumnNotMatch("cannot find PRISM SN column")
    powerSupplyPhaseColumnId = self._searchConfColumnIdsMatch(columns=columns,includeList=["PRISM","supply","phase"])
    if (powerSupplyPhaseColumnId == None): raise MondayProcessorBoardColumnNotMatch("cannot find PRISM power supply phase column")

    ret = [ctColumnId, phaseColumnId, channelColumnId, snColumnId, powerSupplyPhaseColumnId]
    return ret




def plot_the_legend(ax, 
                    fontsize=None,
                    bbox_to_anchor=None):

    if fontsize is None:
        fontsize = 13

    if bbox_to_anchor is None:
        bbox_to_anchor=(0.5, 0.99)

    # plot the legend
    ax.legend()
    legends = ax.get_legend()
    orig_handles = legends.legend_handles[:2]
    # for handle in orig_handles:
    #     print(f"{handle.set(ec='k') = }")

    # adj_handles = [handle.set(ec='k') for handle in orig_handles]
    _ = ax.legend(handles=orig_handles,
                 labels=['Total', 'In Hours'], 
                 loc='upper center', 
                 bbox_to_anchor=bbox_to_anchor, 
                 fontsize=fontsize, 
                 ncol=2)


from . import datalake
from .connect_lake_and_fetch_json import connect_lake_and_fetch_json

def get_conn_by_id_synapse(conn_id, account_key):
    
    credentials = connect_lake_and_fetch_json(datalake, account_key)

    for connection in credentials['connections']:
        if connection['connection_id'] == conn_id:
            return connection['secrets']
    
    print('[INFO]: cannot found the connection_id!')
    return None # if there is no such credential id yet

    # raise('cannot found the connection_id!')



from .send_email import send_email
from .get_group_with_others import get_group_with_others
from .enriching_time_features import enriching_time_features
from .modify_group_names import modify_group_names
from .delete_files_in_directory import delete_files_in_directory
from .color_dict import color_dict
from .generate_time_partition_list import generate_time_partition_list
from .cache_obj import cache_obj
from .setup_table import setup_table
from .setup_table import empty_table
from .patch_missing_periods import patch_missing_periods
from .cache_df_to_db import cache_df_to_db
from .plot_stacked_barchart import plot_stacked_barchart
from .get_pos_bot_range_unit_format import get_pos_bot_range_unit_format






def get_mailing_list(sr_contact_row):
    
    monday_columns = sr_contact_row.index
    email_columns = [column for column in monday_columns if 'email' in column]
    ls_emails_site = sr_contact_row[email_columns].tolist()
    ls_emails_site_nona = [i for i in ls_emails_site if isinstance(i, str)]
    ls_emails_site_nona_with_at = [email_str for email_str in ls_emails_site_nona if '@' in email_str]

    return ls_emails_site_nona_with_at



from .load_mailing_list import load_mailing_list
from .format_contacts import format_contacts

def extract_contacts(sharepoint):

    df_contact_raw = load_mailing_list(sharepoint)

    df_contact = format_contacts(df_contact_raw, 
                                 columns_concerned=sharepoint.excel.columns_concerned
                                 )

    return df_contact


import pickle
from PIL import ImageColor
import os
import io

import colorsys

from mdl.utils_ import color_dict
from .PDF import PDF

cd = color_dict
        
def generate_report(site_name, period, 
                    statements_list=None, 
                    organisation=None, 
                    components=None,
                    files_folder=None, 
                    figures_folder=None, 
                    report_file_folder=None,
                    assets_folder=None):
    
    if assets_folder == None:
        assets_folder = os.path.join(os.path.dirname(__file__), '_assets_/')

    pdf = PDF()
    pdf.add_page()

    pdf.image(assets_folder+'Screenshot_by_date_weekly.png', 0.3, 70, 8)
    pdf.image(assets_folder+'Screenshot_by_asset_weekly.png', 0.3, 70+65, 8)
    pdf.image(assets_folder+'Screenshot_insights_weekly.png', 0.3, 70+130, 8)
    pdf.set_font('Arial', "B", 28)
    # Line break
    pdf.ln(19)

    color_rgb = ImageColor.getcolor(cd["Mindsett Blue"], "RGB")
    color_hsv = colorsys.rgb_to_hsv(*color_rgb)
    color_rgb_changed = colorsys.hsv_to_rgb(color_hsv[0], color_hsv[1]*1.2, color_hsv[2]*0.85)
    pdf.set_text_color(*color_rgb_changed)

    pdf.cell(pdf.w - 28, 10, 'Energy Consumption', 0, 0, 'R')
    pdf.set_font('Arial', "I", 16)
    # Line break
    pdf.ln(13)

    color_rgb = ImageColor.getcolor(cd["darkgrey"], "RGB")
    color_hsv = colorsys.rgb_to_hsv(*color_rgb)
    color_rgb_changed = colorsys.hsv_to_rgb(color_hsv[0], color_hsv[1], color_hsv[2])
    pdf.set_text_color(*color_rgb_changed)

    # generate period name
    
    if period.freqstr == "M":
    
        period_str = period.strftime('%b %Y')
    else:
        # add 1 week to the current_period_obj so that the start week is not Week 00
        period_week_str = period.strftime('%W')
        period_year_str = period.strftime('%Y')
        period_str = f'Wk {(int(period_week_str)+1):02d}, {period_year_str}'
    
    REPLACEMENTS = {"""''""": """'""",
                    "Mecca Bingo": "MB",
                    "Grosvenor": "GC",
                    # "The Rank Group": "Rank",
                    "Marina Bay Financial Center, Tower 2": "MBFC T2"}

    for item in REPLACEMENTS:
        site_name = site_name.replace(item, REPLACEMENTS[item])

    site_name_update = site_name
    sub_title_name = f'{site_name_update} - {period_str} '
    
    if organisation is not None:
        for item in REPLACEMENTS:
            organisation = organisation.replace(item, REPLACEMENTS[item])
        organisation_update = organisation
        sub_title_name = f'{organisation_update} - {sub_title_name}'

    report_file_name = 'Mindsett_Energy_Report_' + sub_title_name.strip().replace(" - ", "_").replace(" ", "_").replace(",", "") + '.pdf'

    pdf.cell(pdf.w - 27, 10, sub_title_name, 0, 0, 'R')

    # if components != None:

    consumption_by_assetclass_piechart_with_phase = components['consumption_by_assetclass_piechart_with_phase.png']
    period_barchart = components['period_barchart.png']
    monthly_total_and_bm_latest = components['monthly_total_and_bm_latest.png']
    daily_barchart_with_temp = components['daily_barchart_with_temp.png']
    auto_statements_list_or_path = components['statements.pkl']

    if isinstance(auto_statements_list_or_path, str):
        with open(auto_statements_list_or_path, 'rb') as f:
            auto_statements_list = pickle.load(f)
    else:
        auto_statements_list = auto_statements_list_or_path
    # else:
    #     consumption_by_assetclass_piechart_mindsett = figures_folder+'consumption_by_assetclass_piechart_mindsett.png'
    #     total_consumption_barchart_with_co2 = figures_folder+'total_consumption_barchart_with_co2.png'
    #     monthly_total_and_bm_latest = figures_folder+'monthly_total_and_bm_latest.png'
    #     daily_consumption_barchart_with_occupancy_mar_with_pattern_mwh = figures_folder+'daily_consumption_barchart_with_occupancy_mar_with_pattern_mwh.png'

    #     with open(files_folder+'statements.pkl', 'rb') as f:
    #         auto_statements_list = pickle.load(f)

    pdf.image(consumption_by_assetclass_piechart_with_phase, 14 + 0.2, 55+69+4, 144)
    pdf.image(period_barchart, 136, 55+82-1, 60)
    pdf.image(monthly_total_and_bm_latest, 14.5, 48+9+3.5, 65-7) # here 3.5 is added to make space for the subtitle

    pdf.ln(151) #Contol on paragraphs 
    pdf.set_x(29)
    pdf.image(daily_barchart_with_temp, 71, 62+3.5, 135) # here 3.5 is added to make space for the subtitle

    # Insights statements 

    if statements_list is not None:

        insight_statements_list = statements_list

        pdf.ln(5)
        pdf.set_x(20)
        pdf.set_font('Arial', 'I', 12)
        pdf.set_text_color(0,0,0)
        pdf.multi_cell(pdf.w - 50, 10, 'Insights: \n')

        text_list = []

        for statement in insight_statements_list:

            statement_text =  [{'text': '- '+statement+' '}]
            text_list += statement_text

        pdf.set_x(29)
        pdf.write_multicell_with_styles(pdf.w-50,6,text_list)

    # Automated Observations statements
    pdf.ln(5)
    pdf.set_x(20)
    pdf.set_font('Arial', 'I', 12)
    pdf.set_text_color(0,0,0)
    pdf.multi_cell(pdf.w - 50, 10, 'Automated Observations: \n')

    text_list = []

    for statement in auto_statements_list:

        statement_text =  [{'text': '- '+statement+' '}]
        text_list += statement_text

    pdf.set_x(29)
    pdf.write_multicell_with_styles(pdf.w-50,6,text_list) #6


    if report_file_folder == None:
        
        pdf_object = io.BytesIO()
        pdf.output(pdf_object, 'F')
        return {report_file_name: pdf_object}
    else:

        report_file_dir = report_file_folder + report_file_name
        pdf.output(report_file_dir, 'F')
        return {report_file_name: report_file_dir}



import uuid
import numpy as np

from cache_return import cache_return

from .load_mailing_list import load_mailing_list
# from ..cache_result import cache_result
from .get_mailing_list import get_mailing_list

@cache_return
def extract_contacts(monday_token, 
                     board_id,
                     columns_concerned=None, 
                     email_prefix=None):
    
    # email_columns = [f'email {idx+1}' for idx in range(max_no_email)]
    # columns = [building_column, *email_columns]
    # print('columns: ', columns)

    df_contact = load_mailing_list(monday_token, board_id,
                                    columns_concerned=columns_concerned, 
                                    email_prefix=email_prefix,
                                    caching=False)

    # # make the names lower case
    df_contact.columns = [column.lower() for column in df_contact.columns]
    df_contact = df_contact.rename(columns={'name': 'item_name'})
    df_contact['building_id'] = df_contact['building_id'].apply(uuid.UUID)

    # df_contact['review'] = df_contact[role_column].str.contains('Review', na=False)
    # df_contact['receive'] = df_contact[role_column].str.contains('Receive', na=False)

    # df_receive = df_contact.loc[df_contact['receive']]
    # df_review = df_contact.loc[df_contact['review']]

    # contacts = {}

    # contacts['receivers'] = df_receive.groupby(name_column)[email_column].apply(list).reset_index().to_dict('records')
    # contacts['reviewers'] = df_review.groupby(name_column)[email_column].apply(list).reset_index().to_dict('records')
    
    # df_contact.publish = (df_contact.publish == 'v')
    # df_contact.testing = (df_contact.testing == 'v')
    # df_contact.occupancy_available = (df_contact.occupancy_available == 'v')
    df_contact.pct_level_tobe_others = df_contact.pct_level_tobe_others.astype('float')
    df_contact.pct_hide = df_contact.pct_hide.replace('', '0').astype('float')
    df_contact.conv_mwh_price = df_contact.conv_mwh_price.replace('', '0').astype('float')
    # df_contact.insight_statements = df_contact.insight_statements.replace('', None)
    df_contact.floor_sqm = df_contact.floor_sqm.replace('', np.NaN).astype('float')

    df_contact['mailing_list'] = df_contact.apply(get_mailing_list, axis=1)

    monday_columns = df_contact.columns
    email_columns = [column for column in monday_columns if 'email' in column]

    df_contact = df_contact.drop(columns=email_columns)

    return df_contact


from typing import List
import requests


from .MondayProcessorExceptions_ import (
    MondayProcessorBoardNotFound,
    MondayProcessorNoColumInBoard,
    MondayProcessorNoItemInGroup,
    MondayProcessorPageNotSupported
)

def _fetchColumnValueByGroup(self, board_id:int, groupId:str, columns:List[str]):
        
    column_ids = []
    
    for column_name in columns:
        for column_id in self._searchColumnIdsByName(board_id, column_name=column_name):
            column_ids.append(column_id)
    
    columns_str = '["' + '","'.join(column_ids) + '"]'
    # columns_index = ["ct", "phase", "channel", "sn", "powerSupplyPhase"]
    # columnsDict = dict(zip(columns,columns_index))
    headers = {"Authorization": self._token,
               "API-Version": '2024-01'}
    data = {"query": f'''query {{
        boards (ids:{board_id}){{
            id
            name
            groups(ids:"{groupId}"){{
                id
                items_page(limit:500){{
                    cursor
                    items{{
                        id
                        name
                        column_values(ids:{columns_str}){{
                            column{{
                                title
                                }}
                            type
                            text
                            ... on BoardRelationValue {{
                                display_value
                            }}
                            ... on LinkValue {{
                                url_text
                            }}
                            ... on CheckboxValue {{
                                checked
                            }}
                        }}
                    }}
                }}

            }}
        }}
    }}'''}
    r = requests.post('https://api.monday.com/v2', headers=headers, json=data)
    data = r.json()
    if 'data' not in data:
        print(data) 
    # print(data) 
    boards = data['data']['boards']
    if(len(boards) <1):
        raise MondayProcessorBoardNotFound(f"Monday board id ({board_id}) not found when perform MondayProcessor._searchConfColumnIds")
    groups = boards[0]['groups']
    if(len(groups) < 1):
        raise MondayProcessorNoColumInBoard(f"No column in board {board_id}")
    items_page = groups[0]['items_page']
    if items_page['cursor'] is not None:
        raise MondayProcessorPageNotSupported(f"The items on page exceeded the limit 500, board {board_id}, group {groupId}!")

    items = items_page['items']
    if(len(items) < 1):
        raise MondayProcessorNoItemInGroup(f"No column in board {board_id}, group {groupId}")
    
    ret = []
    for item in items:
#             print('item:', item)
        tmpDict = {}
        if 'Name' in columns:
            tmpDict['Name'] = item['name']
        for v in item["column_values"]:

            if v["type"] == 'board_relation':
                tmpDict[v['column']["title"]] = v["display_value"]
            elif v["type"] == "link":
                tmpDict[v['column']["title"]] = v["url_text"]
            elif v["type"] == "checkbox":
                tmpDict[v['column']["title"]] = v["checked"]
            else:
                tmpDict[v['column']["title"]] = v["text"]
#                 key = v["id"]
#                 ai = v["additional_info"]
#                 tmpDict[key]= None if ai== None else json.loads(ai)["label"]
#             try:
#                 ch = tmpDict["channel"]
#                 chN = int(ch)
#                 tmpDict["channel"] = chN
#             except:
#                 tmpDict["channel"] = None
        ret.append(tmpDict)
    return ret




def annotate_pct_on_pies(ax, sr, ls_orig_pct_fmt, 
                        pct_hide=None,
                        pct_out=None):
    
    if pct_hide is None:
        pct_hide = 0.04

    if pct_out is None:
        pct_out = 0.065

    for index, pct_value in enumerate(sr.to_list()):

        ax.texts[index*2+1].update({"text": ls_orig_pct_fmt[index]})

        if abs(pct_value) < pct_hide * sr.sum(): 
            ax.texts[index*2+1].update({"text": ''})
        elif abs(pct_value) < pct_out * sr.sum(): 
            ax.texts[index*2].update({"text": f'{ax.texts[index*2+1].get_text()}'})
            ax.texts[index*2+1].update({"text": ''})


import matplotlib.pyplot as plt
import io

from .preprocessing_for_period_barchart import preprocessing_for_period_barchart
from .plot_shifted_period_barchart import plot_shifted_period_barchart


def generate_period_barchart(df_meta_with_value_building,
                            fontsize = None,
                          directory_to_savefig=None):
                          
    df_grouped_working_hours_period_unstacked = preprocessing_for_period_barchart(df_meta_with_value_building)

    # for testing
    # df_grouped_working_hours_period_unstacked[True] = -0.3 * df_grouped_working_hours_period_unstacked[False]
    
    plot_shifted_period_barchart(df_grouped_working_hours_period_unstacked, fontsize=fontsize)

    png_name = 'period_barchart.png'
    if directory_to_savefig == None:
        png_object = io.BytesIO()
        plt.savefig(png_object, format='png', dpi=200)
        plt.close()
        return {png_name: png_object}
    else:
        png_path = directory_to_savefig + png_name
        plt.savefig(png_path, format='png', dpi=200)
        plt.close()
        return {png_name: png_path}




# ruler annotation shinking config

config_dict = {'full_name': {'size_lim': 0.11, 'names': {'part_one': "GOOD", 'part_two': 'NORM', 'part_three': 'POOR'}},
                       'shrink_name':  {'size_lim': 0.08, 'names': {'part_one': "GD", 'part_two': 'NR', 'part_three': 'PR'}},
                       'short_name':  {'size_lim': 0.06, 'names': {'part_one': "G", 'part_two': 'N', 'part_three': 'P'}},
                       'null_name':  {'size_lim': 0.05, 'names': {'part_one': "", 'part_two': '', 'part_three': ''}},
}

def ruler_anno_shink_cf(bar_width, part):

    if bar_width > config_dict['full_name']['size_lim']:
        rpa_text = config_dict['full_name']['names'][part]
    elif bar_width > config_dict['shrink_name']['size_lim']:
        rpa_text = config_dict['shrink_name']['names'][part]
    elif bar_width > config_dict['short_name']['size_lim']:
        rpa_text = config_dict['short_name']['names'][part]
    else:
        rpa_text = config_dict['null_name']['names'][part]
    
    return rpa_text






import pandas as pd

def patch_missing_periods(df_source, start_time, end_time, 
                          time_col='time',
                          value_col='kwh',
                          period_freq='W',
                          timezone='UTC'):
    
    # print('start_time, end_time:', start_time, end_time)

    end_time_inclusive = end_time - pd.Timedelta(microseconds=1)

    # print('df_source: ', df_source.sort_values(by=['time'], ascending=False).head())
    actual_periods_ls = pd.to_datetime(df_source.time, utc=True).dt.tz_convert(tz=timezone).dt.tz_localize(None).dt.to_period(freq=period_freq).unique()
    expect_periods_ls = pd.period_range(start=start_time, end=end_time_inclusive, freq=period_freq)

    # print('actual_periods_ls: ', actual_periods_ls)
    # print('expect_periods_ls: ', expect_periods_ls)

    for period_obj in expect_periods_ls:
        if period_obj not in actual_periods_ls:
            
            print(f'There is no data in period {period_obj}. A placeholder with zero consumption has been added!')
            print('start_time', period_obj.start_time)
            df_placeholder = df_source.head(1)
            with pd.option_context('mode.chained_assignment', None):
                df_placeholder[value_col] = 0
                df_placeholder[time_col] = pd.Timestamp(period_obj.start_time, tz=timezone)
                df_source = pd.concat([df_source, df_placeholder], ignore_index=True)
                df_source.loc[:, time_col] = pd.to_datetime(df_source[time_col], utc=True)
    
    # print('df_source: ', df_source.head())
    return df_source


import pandas as pd

from cache_return import cache_return

@cache_return
def load_building(engine, table_name):

    with engine.connect() as conn:

        query = f""" select id as building_id
                            ,name as building_name
                            ,org_id
                            from {table_name} """ # meta.space_annotations

        df_building = pd.read_sql_query(query, con=conn) 

        df_building.building_name = df_building.building_name.str.strip()
        
    return df_building



import pandas as pd


def get_schedule_records(self, db_schedule):

    # print(f'self.site_name: "{self.site_name}"')
    # print(f'self.org_name: "{self.org_name}"')

    statement_list = [f"""bs.id = '{self.building_id}'"""]

    if self.org_name is not None:
        statement_new  = f"""os.id = '{self.org_id}'"""
        statement_list.append(statement_new)
    statement_full = " and ".join(statement_list)

    engine = db_schedule.engine # new installation - The Hub - ENGINE_WH

    with engine.connect() as conn:
        query = f"""select ss.day
                    ,ss.name
                    ,ss.start_time
                    ,ss.end_time
                    ,BTRIM(bs.name) as site_name 
                    ,BTRIM(os.name) as organisation
                    from {db_schedule.table_schedule.name} ss
                    join {db_schedule.table_building.name} bs on ss.space_id = bs.id
                    join {db_schedule.table_org.name} os on bs.org_id = os.id
                    where {statement_full};"""
        df_schedule = pd.read_sql_query(query,
                                        con=conn) #new installation - The Hub - table_name_wh
        
        # print("schedule_query: ", query)
    return df_schedule.to_dict('records')


import os
import pickle
import inspect
import re
import traceback
# from varname import nameof

def cache_obj(varible, var_name=None, cache_dir='./_cache_', max_count=10000, mode='w', pkg_name=None, print_on=True):

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    
    if mode == 'r':
        if pkg_name is None:
            
            pkl_path = f'{cache_dir}/{varible}.pkl'
        else:
            pkl_path = f'../{pkg_name}_pkg//{cache_dir[1:]}/{varible}.pkl'

        with open(pkl_path, 'rb') as file:
            var_value = pickle.load(file)
            return var_value
    
    if var_name is None: # get the variable name by looping the parent function
        # lcls = inspect.stack()[2][0].f_locals
        # for name in lcls:
        #     if id(varible) == id(lcls[name]):
        #         var_name = name

        stack = traceback.extract_stack()
        filename, lineno, function_name, code = stack[-2]
        var_name_raw = re.compile(r'\((.*?)\).*$').search(code).groups()[0] # the default name is everything in the bracket, like: 'None_0', mode='w'
        var_name = var_name_raw.split(',')[0] 
        pkl_path = f'{cache_dir}/{var_name}.pkl'

    if var_name is None: # get the name by adding incremental number on the default name
        # dir_path = os.path.dirname(os.path.realpath(__file__))
        var_name = 'cached_variable'
        for count_of_cached_obj in range(max_count):
            pkl_path = f'{cache_dir}/{var_name}_{count_of_cached_obj}.pkl'
            if not os.path.exists(pkl_path):
                break
        
    with open(pkl_path, 'wb') as file:
        pickle.dump(varible, file)

    if print_on:
        print(var_name+": ", varible)

    print(f'[INFO]: the value of variable [{var_name}] has been cached under ./_cache_!')




from sqlalchemy import create_engine

from get_conn_by_id import get_conn_by_id 

from . import table_cache
from . import table_mailing
from . import table_iot
from . import table_meta
from . import table_schedule
from . import table_building
from . import table_org
from . import table_timezone
from . import table_sqm
from . import table_price
from . import table_on_off
from . import table_building_total
from . import table_occupancy

# credentials/setup for postgresdb

postgresdb_conn = get_conn_by_id('prod_postgres_cloudfm')

DB_URL = postgresdb_conn['host']+":"+str(postgresdb_conn['port'])
DB_DATABASE = postgresdb_conn['database']
DB_USER = postgresdb_conn['username']
DB_PASSWORD = postgresdb_conn['password']
DB_SSLMODE = postgresdb_conn["sslmode"]
ENGINE = f"postgresql://{DB_USER}:{DB_PASSWORD}@{DB_URL}/{DB_DATABASE}"

engine = create_engine(ENGINE)

# columns_for_query = ['circuit_description', 'asset_type', 'asset_class', 'asset_location', 'phase', 'channel_number', 'nid', 'site_name', 'organisation']




def check_and_keep_valid(df_config):

    df_config = df_config.fillna('')

    # df_config = keep_valid(df_config, columns=['industry', 'floor_size', 'conv_mwh_price', 'timezone'])

    no_industry = (df_config.industry == '')
    no_floor_size = (df_config.floor_size.isna())

    # print('df_config.conv_mwh_price: ', df_config.conv_mwh_price)

    no_conv_mwh_price = (df_config.conv_mwh_price == 0)
    no_timezone = (df_config.timezone.isna())

    not_ready = (no_industry | no_floor_size | no_conv_mwh_price | no_timezone)

    to_report = (df_config.publish | df_config.testing)

    not_ready_but_report = (not_ready & to_report)

    df_not_ready_but_report = df_config.loc[not_ready_but_report]

    if df_not_ready_but_report.shape[0] > 0:
        
        print('df_not_ready_but_report.T: ', df_not_ready_but_report.T)

        raise Exception('df_not_ready_but_report is not empty, please double check the config of [industry, floor_size, conv_mwh_price, timezone] for these sites!')
    
    df_config_ready = df_config.loc[~not_ready]

    # quit()

    return df_config_ready





import smtplib
from pathlib import Path
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from email import encoders
import io

def send_email(send_from, send_to, subject, message, signature=None, cc=None, bcc=None, files={},
              server="localhost", port=587, username='', password='',
              use_tls=True):
    """Compose and send email with provided info and attachments.

    Args:
        send_from (str): from name
        send_to (list[str]): to name(s)
        subject (str): message title
        message (str): message body
        files (dict{str:str|obj}): list of file paths to be attached to email
        server (str): mail server host name
        port (int): port number
        username (str): server auth username
        password (str): server auth password
        use_tls (bool): use TLS mode
    """
    
    msg = MIMEMultipart()
    msg['From'] = send_from
    msg['To'] = COMMASPACE.join(send_to)
    
    if cc is not None:
        msg['Cc'] = COMMASPACE.join(cc)
        
    if bcc is not None:
        msg['Bcc'] = COMMASPACE.join(bcc)
        
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = subject

    html_msg = MIMEText(message, 'html')
    msg.attach(html_msg)

    if signature != None:
        sgn = MIMEText(signature, "html")
        msg.attach(sgn)

    for filename in files:
        part = MIMEBase('application', "octet-stream")

        file_path_obj = files[filename]

        if isinstance(file_path_obj, io.BytesIO):
            part.set_payload(file_path_obj.getvalue())
            encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            f'attachment; filename={filename}')
        else:
            with open(file_path_obj, 'rb') as file:
                part.set_payload(file.read())
            encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            f'attachment; filename={filename}')
        msg.attach(part)

    smtp = smtplib.SMTP(server, port)
    if use_tls:
        smtp.starttls()
    smtp.login(username, password)
    smtp.sendmail(send_from, send_to, msg.as_string())
    smtp.quit()
     
    print(f'[smtp]: email sent successfully to {str(send_to)}!')


import pandas as pd

from .load_metadata import load_metadata
from .load_iot_data import load_iot_data

def import_data_with_meta_postgres(db, start_time, end_time, building_id, 
                        #   organisation=None, 
                          exception=None,
                          chunksize=10,
                          # meta_columns_for_join=['nid', 'channel_number'],
                          # iot_columns_for_join=['nid', 'channel'],
                          # reading_interval_in_mins=10,
                          timezone='UTC'):  # the resampling operation should be implemented later

    # format the time range
    query_start_time = pd.Timestamp(start_time, tz=timezone)
    query_end_time  =  pd.Timestamp(end_time, tz=timezone)

    query_start_time_str = query_start_time.strftime("%Y-%m-%d %H:%M:%S.%f%z")
    query_end_time_str   = query_end_time.strftime("%Y-%m-%d %H:%M:%S.%f%z")

    time_range = [query_start_time_str, query_end_time_str]

    engine = db.engine
    table_name_iot = db.table_iot.name
    table_name_meta = db.table_meta.name
    # table_name_link = db.table_name_link

    df_meta = load_metadata(engine, table_name_meta, building_id, 
                            # organisation=organisation,
                            exception=exception)

    if df_meta.shape[0] == 0:

        raise Exception('There is not metadata found for this building!')

    # print('df_meta: ', df_meta)

    id_list = df_meta.id.to_list()

    df_emd = load_iot_data(engine, time_range, table_name_iot, id_list, chunksize=chunksize)

    df_meta_with_value = df_emd.merge(df_meta, on='id').drop(columns=['id']).rename(columns={'w': 'W'})

    # print('df_meta_with_value.shape: ', df_meta_with_value.shape)

    return df_meta_with_value


import pandas as pd

def generate_time_partition_list(time_range, freq='1D'):
    
    start_time_str = time_range[0]
    end_time_str = time_range[1]
    
    time_list = pd.date_range(start=start_time_str, end=end_time_str, freq=freq).to_list()

    time_str_list = [time.strftime("%Y-%m-%d %X%z") for time in time_list]

    time_range_list = list(zip(time_str_list[:-1], time_str_list[1:]))
    
    return time_range_list



from fetch_sharepoint_info import fetch_sharepoint_info

from .check_and_keep_valid import check_and_keep_valid
from .load_timezone import load_timezone
from .load_sqm import load_sqm
from .attach_id_to_buildings import attach_id_to_buildings
from .apply_price_from_db import apply_price_from_db

def load_config(engine, 
                sharepoint,
                table_name_org,
                table_name_building,
                table_name_timezone, 
                table_name_sqm,
                table_name_price,
                table_name_cache, 
                table_columns_cache,
                monday=None, # making if optional here, so in the future it can be ignored from the outer layer of functions
                sync_source=None,
                cfg_src_type=None):
    
    if cfg_src_type is None:
         cfg_src_type = 'sharepoint' # default to sharepoint as suggested by Pat


    if cfg_src_type == 'sharepoint':
         
         df_config_raw = fetch_sharepoint_info(engine, 
                                                sharepoint, 
                                                table_name_cache, 
                                                table_columns_cache,
                                                sync_source=sync_source)
    
    elif cfg_src_type == 'monday': 

        from fetch_monday_info import fetch_monday_info

        if monday is None:
             raise Exception('[ERROR]: the config for monday is not provided, but cfg_src_type is monday!')

        df_config_raw = fetch_monday_info(engine, 
                                    monday.auth.TOKEN,
                                    table_name_cache, 
                                    table_columns_cache,
                                    monday.board_id,
                                    columns_concerned=monday.columns_concerned, 
                                    email_prefix=monday.email_prefix,
                                    sync_monday=sync_source)
    
    else:
        raise Exception('[ERROR]: please choose the cfg_src_type from options [sharepoint, monday].')

        
    df_config = attach_id_to_buildings(engine, df_config_raw, table_name_org, table_name_building, 
                                       caching=False) # df_config has to be the same name, so not renaming it here

    df_tz = load_timezone(engine, table_name_timezone)
    df_config_tz = df_config.merge(df_tz, on='building_id', how='left')

    df_sqm = load_sqm(engine, table_name_sqm)
    df_config_tz_sqm = df_config_tz.merge(df_sqm, on='building_id', how='left')

    df_config_tz_price = apply_price_from_db(engine, table_name_price, df_config_tz_sqm)
    # df_price = load_energy_price(engine, table_name_price)
    # df_config_tz_sqm = df_config_tz_sqm.rename(columns={'conv_mwh_price': 'conv_mwh_price_bak'})
    # # df_config_tz_sqm_no_price = df_config_tz_sqm.drop(columns=['conv_mwh_price'])
    # df_config_tz_price = df_config_tz_sqm.merge(df_price, on='building_id', how='left')
    # df_config_tz_price['conv_mwh_price'] = df_config_tz_price['conv_mwh_price'].fillna(df_config_tz_price['conv_mwh_price_bak'])
    # df_config_tz_price = df_config_tz_price.drop(columns=['conv_mwh_price_bak'])
    
    df_config_vld = check_and_keep_valid(df_config_tz_price)
    
    return df_config_vld
        


def create_new_schema(engine, schema_name):
    
    with engine.connect() as conn:
        query_create_schema = f"""CREATE SCHEMA IF NOT EXISTS {schema_name};"""
        conn.execute(query_create_schema) 
    
    print(f'New schema [{schema_name}] is created.')
    return



from azure.storage.filedatalake import DataLakeServiceClient
import io
import json


def connect_lake_and_fetch_json(datalake, account_key):

    # connect to the datalake and fetch the json file
    azure_service_client = DataLakeServiceClient(account_url="{}://{}.dfs.core.windows.net".format(
            "https", datalake.account_name), credential=account_key)
        
    azure_file_system_client = azure_service_client.get_file_system_client(file_system=datalake.container_name)

    file_client = azure_file_system_client.get_file_client(datalake.file_path)
    data = file_client.download_file().readall()
    json_dict = json.load(io.BytesIO(data))

    return json_dict



# import matplotlib.patches as mpatches

from .plot_phase_balance_part import plot_phase_balance_part
from .get_tick_values import get_tick_values
from .get_phase_markers import get_phase_markers

color_dict = {"grey":  '#808080',
              "brown": '#964B00',
              "black": '#000000'}

def plot_phase_balance(ax, phase_values_list,
                    fontsize=None,
                    ru_x=None,
                    ru_y=None,
                    ru_h=None,
                    ru_w=None):
    
    phase_markers_list = get_phase_markers(phase_values_list)

    part_control_ls = [{'ruler_part_color': color_dict["brown"],
                        'phase_marker': phase_markers_list[0],
                         'part': 'part_one',
                         'part_idx': 1},
                        {'ruler_part_color': color_dict["black"],
                        'phase_marker': phase_markers_list[1],
                         'part': 'part_two',
                         'part_idx': 2},
                        {'ruler_part_color': color_dict["grey"],
                        'phase_marker': phase_markers_list[2],
                         'part': 'part_three',
                         'part_idx': 3}]
    
    ticks_values_list = get_tick_values(phase_values_list)

    for part_control in part_control_ls:
        plot_phase_balance_part(ax, ticks_values_list, part_control,
                                fontsize=fontsize,
                                ru_x=ru_x,
                                ru_y=ru_y,
                                ru_h=ru_h,
                                ru_w=ru_w)




import matplotlib.pyplot as plt

from os import path

from .energy_meter_design import energy_meter_design
from .benchmarking_design import benchmarking_design


def energy_meter_with_benchmarking(consumption_mwh_cur, consumption_mwh_pre, floor_size, 
                                   image_folder=None,
                                    size_in_sqm = True,
                                    industry = "office",
                                    period = 30,
                                    conv_mwh_co2 = 0.233,
                                    conv_mwh_price = 190,
                                    fx = 0.2, # fancy box x-axis value
                                    fy = 0.3, # fancy box y-axis value
                                    fs  = 0.6, # fancy box scale factor
                                    kwh_per_sqm_good = 10.33,
                                    kwh_per_sqm_typical = 14.5,
                                    currency = 'GBP',
                                    fontsize=13, # please note that the other fontsizes haven't been tested yet
                                    # directory_to_savefig = './figures/'
                                  ):
    
    if image_folder == None:
        image_folder = path.join(path.dirname(__file__), '_assets_/')

    plt.style.use('seaborn-white')
    fig, ax = plt.subplots(figsize=(5, 5.5))

    if size_in_sqm:
        floor_sqm = floor_size
    else:
        floor_sqm = floor_size * 0.0929

    if industry == "office":
        floor_sqm_as_office = floor_sqm
    elif industry == "food service":
        # (not working on 26 Nov 24): https://cdn2.hubspot.net/hubfs/5055051/Offers/Energy%20Benchmarking%20Report%20-%20Iota.pdf?utm_campaign=Offer%3A%20Energy%20Benchmarking%20Report&utm_medium=email&_hsmi=72631957&_hsenc=p2ANqtz-8urx_6ejMPQ25rp-u0vAHPq0cmKPTvL18SQTEf22gtrdDV2x7wGnd5kkP40_bx3M5hOWp3tysnbPI4JjWriJEp2fb5o7PzNF5D9VFqQNjYVLVxKtE&utm_content=72631957&utm_source=hs_automation
        # added on 26 Nov 24: https://www.betterbuildingspartnership.co.uk/sites/default/files/media/attachment/REEB%202023%20Benchmarks_0.pdf
        floor_sqm_as_office = floor_sqm * 56/15 # 56/15 food service/office

    floor_sqm_as_office_for_month = floor_sqm_as_office * period / 30 # default days in a month

    
    energy_meter_design(ax, consumption_mwh_cur, consumption_mwh_pre, image_folder,
                        conv_mwh_co2=conv_mwh_co2,
                        conv_mwh_price=conv_mwh_price,
                        fx=fx, fy=fy, fs=fs,
                        currency=currency,
                        fontsize=fontsize)
    benchmarking_design(ax, consumption_mwh_cur, floor_sqm_as_office_for_month,
                        fx=fx, fy=fy-0.12, fs=fs,
                        kwh_per_sqm_good=kwh_per_sqm_good,
                        kwh_per_sqm_typical=kwh_per_sqm_typical,
                        fontsize=fontsize
                       )
    ax.set_xlim(0, 1) # shouldn't be modified
    ax.set_ylim(-1, 2)

    plt.axis('off')
    # Specify the directory to save figures, if it does not exist, create it
    # Path(directory_to_savefig).mkdir(parents=True, exist_ok=True)
    # plt.savefig(directory_to_savefig+"Monthly_total_and_bm_latest.png", format='png', dpi=200,transparent=True, bbox_inches='tight', pad_inches=0)




import matplotlib.patches as mpatches

from .ruler_anno_shink_cf import ruler_anno_shink_cf

def plot_phase_balance_part(ax, ticks_values_list, part_control,
                         fontsize=None,
                         fx=None,
                         fy=None,
                         fs=None,
                         ru_x=None,
                         ru_y=None,
                         ru_h=None,
                         ru_w=None):

    if fontsize is None:
        fontsize = 13

    # fancy box setting
    if fx is None:
        fx = -1  # fancy box x-axis value
    if fy is None:
        fy = -1.3  # fancy box y-axis value
    if fs is None:
        fs = 1  # fancy box scale factor

    # ruler settings
    if ru_x is None:
        ru_x = 0  # the x-axis of benchmarking ruler
    if ru_y is None:
        ru_y = 0  # the x-axis of benchmarking ruler
    if ru_h is None:
        ru_h = 0.03  # the height of benchmarking ruler
    if ru_w is None:
        ru_w = 1  # the width of benchmarking ruler

    ruler_part_color = part_control['ruler_part_color']
    part = part_control['part']
    part_idx = part_control['part_idx']
    phase_marker = part_control['phase_marker']
    # text_color = part_control['text_color']

    # ticks range
    ticks_max = max(ticks_values_list)
    ticks_min = min(ticks_values_list)
    ticks_range = ticks_max - ticks_min
    
    # rp_edgecolor = 'k' # edgecolor of the ruler body parts
    rp_note_fontsize = fontsize - 5 # the fontsize of notes on ruler parts # please note that the other sizes except default 13 haven't been tested yet

    # ruler body - rectangle bar
    rp_x = ru_x + ru_w * (ticks_values_list[part_idx-1]-ticks_min)/ticks_range       # the x-axis value of this part
    rp_y = ru_y - 0.03 # - 0.1 # 0.14      # the y-axis value of this part
    rp_pct = (ticks_values_list[part_idx]-ticks_values_list[part_idx-1])/ticks_range
    rp_w = ru_w * rp_pct   # the width value of this part
    rp_h = ru_h
    rp_c = ruler_part_color
    ruler_body_part = mpatches.Rectangle([fx+rp_x*fs, fy+rp_y*fs], rp_w*fs, rp_h*fs, 
                                         facecolor=rp_c, 
                                         edgecolor='w', 
                                         lw=2)
    ax.add_artist(ruler_body_part)

    # ruler body - annotation on the bar
    rpa_x = rp_x + rp_w/2.0 # ruler body part annotation - x axis
    rpa_y = rp_y + 0.076 # ruler body part annotation - y axis
    bar_width = rp_w*fs
    # part = 'part_one'
    rpa_text = ruler_anno_shink_cf(bar_width, part)
    
    ax.annotate(rpa_text, (fx+fs*rpa_x, fy+fs*rpa_y), 
                color= 'w', # '#B87333', # text_color, 
                weight='extra bold', 
                fontsize=rp_note_fontsize, 
                ha='center', 
                va='center')
    
    # ruler body - marker symbol below the bar
    rps_x = rp_x + rp_w/2.0 # ruler body part annotation - x axis
    sym_pad_y = (-0.18)
    rps_y = rp_y + sym_pad_y  # ruler body part annotation - y axis
    bar_width = rp_w*fs
    rp_maker_fontsize = fontsize +2 # make it bigger than the annotation text, to be more visible

    # print(f'{rp_pct = }')

    # balance_value = 1./3
    # variance_allowance = 0.04/3

    # if rp_pct > (balance_value + variance_allowance):
    #     rps_text = '>'
    # elif rp_pct > (balance_value - variance_allowance):
    #     rps_text = '='
    # elif rp_pct > 0:
    #     rps_text = '<'
    # else:
    #     rps_text = '?'

    # rpa_text = ruler_anno_shink_cf(bar_width, part)
    symbol_color = 'k'
    
    ax.annotate(phase_marker, (fx+fs*rps_x, fy+fs*rps_y), 
                color=symbol_color, 
                weight='extra bold', 
                fontsize=rp_maker_fontsize, 
                ha='center', 
                va='center')
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#author: Xiao Wang 
# x.wang@cloudfmgroup.com
# Created: 05/04/2023

from .DataProcessor import DataProcessor
from monday import MondayClient

class MondayProcessor(DataProcessor):

    from .launchInteractiveProcess import launchInteractiveProcess
    from ._fetchBoardById import _fetchBoardById
    from ._fetchBoardLogsById import _fetchBoardLogsById
    from ._fetchGroups import _fetchGroups
    from ._fetchPrismConf import _fetchPrismConf
    from ._fetchPrismConfByBoard import _fetchPrismConfByBoard
    from ._fetchPrismConfByGroup import _fetchPrismConfByGroup
    from ._fetchColumnValueByGroup import _fetchColumnValueByGroup
    from ._searchConfColumnIds import _searchConfColumnIds
    from ._searchColumnIdsByName import _searchColumnIdsByName
    from ._searchConfColumnIdsMatch import _searchConfColumnIdsMatch
    
    def __init__(self, MONDAY_TOKEN) -> None:
        super().__init__()
        self._token = MONDAY_TOKEN
        self.monday = MondayClient(self._token)
    

if __name__ == "__main__":

    mp = MondayProcessor()
    # mp._fetchBoardById(4246692749)
    # mp._fetchPrismConfByGroup(board_id=4246692749, groupId="topics", columns=["ct_type",  "color",  "prism_channel_no_9", "status1" ])
    ConfColumnIds = mp._searchConfColumnIds(4246692749)
    mp._fetchPrismConfByGroup(board_id=4246692749, groupId="topics", columns=ConfColumnIds)
    # mp._fetchGroups(4246692749)
    # mp._fetchPrismConf(4246692749)
    # df = mp.launchInteractiveProcess(4246692749)
    # print(df)



import pandas as pd

def patch_missing_dates(df_pivot_working_hours, period_current):

    start_date = period_current.start_time.date()
    end_date = period_current.end_time.date()

    # firstly ensuring that the start and finish date are not missing
    if start_date not in df_pivot_working_hours.index:
        df_pivot_working_hours.loc[start_date] = 0

    if end_date not in df_pivot_working_hours.index:
        df_pivot_working_hours.loc[end_date] = 0

    # convert the date index to full datetime, so it can be used to adjust frequency
    df_pivot_working_hours.index = pd.to_datetime(df_pivot_working_hours.index)
    df_pivot_working_hours_patched = df_pivot_working_hours.asfreq('1D', fill_value=0).fillna(0)

    # convert the full datetime frequency to date following the original format
    df_pivot_working_hours_patched.index = df_pivot_working_hours_patched.index.date

    return df_pivot_working_hours_patched



    
    





color_dict = {"tomato": '#FF836A',
      "aquablue": '#6DC2B3',
      "peach": '#FED6D2',
      "darkgrey": '#9F9D9C',
      "potato": '#FEF8C8',
      "cyan": '#B6E4E1',
     "Mindsett Blue": '#132F57',
     "Black": '#000000',
     "Mindsett Grey":'#b3b3b3',
     "White": '#ffffff',
     "Mindsett Green": '#00A19A',
     "Mindsett Red":'#E6354C'}


# from pathlib import Path
from os import path
import pandas as pd
from cache_return import cache_return

from mdl.utils_ import enriching_time_features
from mdl.utils_ import patch_missing_periods

# from .unify_db_interface import unify_db_interface
from .load_source_data import load_source_data

from .validate_data_not_empty import validate_data_not_empty
from .refine_group_name import refine_group_name
# from .patch_missing_periods import patch_missing_periods

@cache_return
def preparing_source_data(postgresdb, 
                          site_obj,
                          debug=False,
                          chunksize=10):
    
    # get the specific time for the individual site
    start_time = site_obj.get_start_time()
    start_time_co2_barchart = site_obj.get_start_time_co2_barchart()
    end_time = site_obj.get_end_time()
    schedule_records = site_obj.get_schedule_records(postgresdb)

    df_meta_with_value, df_meta_with_value_building = load_source_data(postgresdb, 
                                                                        start_time, 
                                                                        start_time_co2_barchart,
                                                                        end_time, 
                                                                        site_obj.building_id,
                                                                        chunksize=chunksize,
                                                                        exception=site_obj.exception,
                                                                        timezone=site_obj.timezone,
                                                                        caching=debug)
    
    # print('df_meta_with_value.shape: ', df_meta_with_value.shape)
    validate_data_not_empty(df_meta_with_value, df_meta_with_value_building)

    # print('df_meta_with_value.info: ', df_meta_with_value.info())

    df_meta_with_value = patch_missing_periods(df_meta_with_value, start_time, end_time,
                                                        time_col='time',
                                                        value_col='W',
                                                        period_freq=site_obj.period_freq,
                                                        timezone=site_obj.timezone)

    df_meta_with_value_building = patch_missing_periods(df_meta_with_value_building, start_time_co2_barchart, end_time,
                                                        time_col='time',
                                                        value_col='kwh',
                                                        period_freq=site_obj.period_freq,
                                                        timezone=site_obj.timezone)
    
    df_meta_with_value = refine_group_name(df_meta_with_value, 
                                           site_obj.asset_group, 
                                           site_obj.group_name_column_exchange, 
                                           site_obj.fillna_value,
                                           site_obj.group_name_modification)

    df_meta_with_value = enriching_time_features(df_meta_with_value, 
                                                    schedule_records= schedule_records,
                                                    period_freq=site_obj.period_freq,
                                                    weekend=site_obj.weekend, 
                                                    working_end_time=site_obj.working_end_time, 
                                                    working_start_time=site_obj.working_start_time,
                                                    timezone=site_obj.timezone)

    df_meta_with_value_building = enriching_time_features(df_meta_with_value_building,
                                                    schedule_records= schedule_records,
                                                    period_freq=site_obj.period_freq,
                                                    weekend=site_obj.weekend, 
                                                    working_end_time=site_obj.working_end_time, 
                                                    working_start_time=site_obj.working_start_time,
                                                    timezone=site_obj.timezone)
    
    # print('df_meta_with_value.shape: ', df_meta_with_value.shape)
    return df_meta_with_value, df_meta_with_value_building



import matplotlib.ticker as ticker

from .generate_day_code import generate_day_code

def setup_x_tick_lim(ax, df_pivot_working_hours_sorted, 
                           fontsize=None):

    if fontsize is None:
        fontsize = 13

    day_code_list = generate_day_code(df_pivot_working_hours_sorted)
    df_pivot_working_hours_sorted_reset = df_pivot_working_hours_sorted.reset_index(drop=True)
        
    # set up the x-ticks and xlim
    ax.tick_params(axis='x', which='major', pad=8, length=2, labelsize=fontsize)

    x_tick_pad =  1.5
    top_index = df_pivot_working_hours_sorted_reset.index.min() - x_tick_pad
    bot_index = df_pivot_working_hours_sorted_reset.index.max() + x_tick_pad
    ax.set_xlim([top_index, bot_index])

    # fixing xticks with matplotlib.ticker "FixedLocator": UserWarning: FixedFormatter should only be used together with FixedLocator
    ticks_loc = ax.get_xticks().tolist()
    ax.xaxis.set_major_locator(ticker.FixedLocator(ticks_loc))
    ax.set_xticklabels(day_code_list, fontsize=fontsize)



from get_conn_by_id import get_conn_by_id 

from . import excel

sharepoint_conn = get_conn_by_id('mindsett_sharepoint')

client_id = sharepoint_conn['client_id']
client_secret = sharepoint_conn['secret']
site_url = sharepoint_conn['host']

folder_url = "/sites/Mindsett/Shared Documents/9. Data Science/Mailing List - Energy Report"


import pandas as pd

from .generate_day_code_list import generate_day_code_list


def generate_day_code(df_pivot_working_hours):
    
    datetime_list = list(pd.to_datetime(df_pivot_working_hours.index))

    day_code_list = generate_day_code_list(datetime_list)

    day_code_list.insert(0,"") # for making spaces at the start
    day_code_list.insert(0,"")
    # day_code_list.insert(0,"")
    day_code_list.append("") # for making spaces at the end
    day_code_list.append("")

    return day_code_list



import matplotlib.ticker as ticker

from .get_pos_bot_and_ylim import get_pos_bot_and_ylim

def setup_y_label_and_get_pos_bot_range(ax, df_pivot_working_hours_sorted, 
                                        fontsize=None):

    if fontsize is None:
        fontsize = 13

    df_pivot_working_hours_sorted_reset = df_pivot_working_hours_sorted.reset_index(drop=True)

    df_pivot_working_hours_sorted_reset_pos, \
        df_pivot_working_hours_sorted_reset_bot, \
            ytick_range = get_pos_bot_and_ylim(df_pivot_working_hours_sorted_reset,
                                    pad_pct_above_max = 0.3)

    max_y_tick = max([abs(y_tick) for y_tick in ytick_range])

    if max_y_tick > 0.1: # 100 kwh = 0.1 mwh as switching point for unit as kwh or mwh
        ax.set_ylabel("Daily Usage (MWh)", labelpad=14, fontsize=fontsize)
        
        #ax_l.set_yticks(np.arange(0, tick_range_e, 0.1))
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1f'))
    else:
        df_pivot_working_hours_sorted_reset = df_pivot_working_hours_sorted_reset.mul(1000)

        # after the change of units, the dfs and limits should be updated
        df_pivot_working_hours_sorted_reset_pos, \
            df_pivot_working_hours_sorted_reset_bot, \
                ytick_range = get_pos_bot_and_ylim(df_pivot_working_hours_sorted_reset,
                                    pad_pct_above_max = 0.3)

        ax.set_ylabel("Daily Usage (kWh)", labelpad=14, fontsize=fontsize)
        
        #ax_l.set_yticks(np.arange(0, tick_range_e, 0.1))
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.0f'))

    return df_pivot_working_hours_sorted_reset_pos, df_pivot_working_hours_sorted_reset_bot, ytick_range


import json
from .get_credential_url import get_credential_url

def get_conn_by_id_json(conn_id, 
                        credential_url=None):
    
    if credential_url == None:
        credential_url = get_credential_url()

        if credential_url == None: # if there is no credential_url from localhost or databricks
            return None

    with open(credential_url) as f:
        credentials = json.load(f)


        
    for connection in credentials['connections']:
        if connection['connection_id'] == conn_id:
            return connection['secrets']
    
    print('[INFO]: cannot found the connection_id!')
    return None # if there is no such credential id yet

    # raise('cannot found the connection_id!')




def generate_pct_fmt(sr):
    total_before_adding_gap = sr.sum()
    ls_orig_pct = (sr/total_before_adding_gap).to_list()
    ls_orig_pct_fmt = ['{:.0f}%'.format(round(p*100)) for p in ls_orig_pct]
    ls_orig_pct_fmt.append('') # add for potential gap value
    return ls_orig_pct_fmt



import uuid

from .uuid_for_nona import uuid_for_nona
from .extract_id import extract_id
from .get_mailing_list import get_mailing_list

def format_contacts(df_contact, 
                    columns_concerned=None):

    if columns_concerned is None:
        from .columns_concerned import columns_concerned

    pct_level_tobe_others_default = 0.03
    pct_hide_default = 4
    conv_mwh_price_default = 190

    # standardize the column names
    df_contact.columns = [column.lower() for column in df_contact.columns]
    df_contact = df_contact.rename(columns={'building_name': 'item_name', 'org': 'group_name'})

    # fill the names for null values
    df_contact['group_name'] = df_contact['group_name'].ffill()
    df_contact['insight_statements'] = df_contact['insight_statements'].fillna('')
    df_contact['group_name_column_exchange'] = df_contact['group_name_column_exchange'].fillna('')
    df_contact['group_name_modification'] = df_contact['group_name_modification'].fillna('')
    df_contact['manager_name'] = df_contact['manager_name'].fillna('')

    # columns type formatting
    df_contact['publish'] = (df_contact['publish'] == 'v')
    df_contact['testing'] = (df_contact['testing'] == 'v')
    df_contact['occupancy_available'] = (df_contact['occupancy_available'] == 'v')
    df_contact['building_id'] = df_contact['building_id'].apply(uuid_for_nona)

    df_contact['pct_level_tobe_others'] = df_contact['pct_level_tobe_others'].fillna(pct_level_tobe_others_default).astype('float')
    df_contact['floor_sqm'] = df_contact['floor_sqm'].astype('float')

    df_contact['pct_hide'] = df_contact['pct_hide'].fillna(pct_hide_default).astype('float')

    df_contact['conv_mwh_price'] = df_contact['conv_mwh_price'].fillna(conv_mwh_price_default).astype('float')

    # format the monday board ids
    df_contact['board_id'] = df_contact['board_id'].apply(extract_id)
    df_contact['hvac_board_id'] = df_contact['hvac_board_id'].apply(extract_id)

    # get_mailing_list
    df_contact['mailing_list'] = df_contact.apply(get_mailing_list, axis=1)

    # drop the email related columns
    monday_columns = df_contact.columns
    email_columns = [column for column in monday_columns if 'email' in column]
    df_contact = df_contact.drop(columns=email_columns)

    # select the concerned columns
    df_contact_sel = df_contact[columns_concerned]

    return df_contact_sel


import matplotlib.ticker as ticker

def set_x_tick_and_lim(ax, x_ticks_labels, df_pivot_working_hours_sorted,
                       fontsize=None):

    if fontsize is None:
        fontsize = 13

    df_pivot_working_hours_sorted_reset = df_pivot_working_hours_sorted.reset_index(drop=True)

    # set up the x-ticks and xlim
    ax.tick_params(axis='x', which='major', pad=8, length=2, labelsize=fontsize)

    x_tick_pad =  0.99
    top_index = df_pivot_working_hours_sorted_reset.index.min() - x_tick_pad
    bot_index = df_pivot_working_hours_sorted_reset.index.max() + x_tick_pad
    ax.set_xlim([top_index, bot_index])

    # used to handle the below warning:
    ax.xaxis.set_major_locator(ticker.MultipleLocator(1)) # ValueError: The number of FixedLocator locations (5), usually from a call to set_ticks, does not match the number of labels (8).

    # fixing yticks with matplotlib.ticker "FixedLocator": UserWarning: FixedFormatter should only be used together with FixedLocator
    ticks_loc = ax.get_xticks().tolist()
    # print(f'{ticks_loc = }')
    ax.xaxis.set_major_locator(ticker.FixedLocator(ticks_loc))
    ax.set_xticklabels(x_ticks_labels, fontsize=fontsize)   


#The below is the table definition
name = 'analysis.energy_report_config'

columns_dict = {'last_update': 'timestamptz NOT NULL', 
                'building_id': 'UUID NOT NULL',
                'item_name': 'text NOT NULL', 
                'group_name': 'text',
                'board_id': 'text',
                'hvac_board_id': 'TEXT',
                'publish': 'boolean NOT NULL', 
                'testing': 'boolean NOT NULL', 
                'industry': 'text', 
                'manager_name': 'text NOT NULL', 
                'currency': 'text NOT NULL', 
                # 'timezone': 'text NOT NULL', 
                'occupancy_available': 'boolean NOT NULL', 
                'asset_group': 'text NOT NULL', 
                'fillna_value': 'text NOT NULL', 
                'insight_statements': 'text NOT NULL', 
                'group_name_column_exchange': 'text NOT NULL', 
                'group_name_modification': 'text NOT NULL', 
                'pct_level_tobe_others': 'double precision NOT NULL', 
                'floor_sqm': 'double precision', 
                'pct_hide': 'double precision NOT NULL', 
                'conv_mwh_price': 'double precision',
                'period_freq': 'text NOT NULL',
                'mailing_list': 'text NOT NULL'}

hypertable = False 






def get_legend_handles(ax, no_of_labels=50):

    ax.legend(labels=['']*no_of_labels)

    legends = ax.get_legend()

    valid_handles = []

    for handle in legends.legend_handles:

        is_invisible_pie_for_gap = (handle.get_visible()==False)
        is_white_pie_in_ring_centre = (handle.get_fc() == (1.0, 1.0, 1.0, 1))
        is_outer_gap_edge_when_as_inner = (handle.get_fc()[3]==0)

        invalid_handle = (is_invisible_pie_for_gap or is_white_pie_in_ring_centre or is_outer_gap_edge_when_as_inner)

        if not invalid_handle: 
            valid_handles.append(handle)

    return valid_handles



import functools
import pandas as pd
import os
import pickle

# note: this function require the cached module function names to be unique
# note: this function can be only applied to functions returning dataframe now 

def cache_return(_func=None):
    def decorator(func):
        func_name = func.__name__
        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            caching = kwargs.pop('caching', False)
            flushing = kwargs.pop('flushing', False)
            cache_path = kwargs.pop('cache_path', './_cache_')

            if caching:
                if not os.path.exists(cache_path):
                    os.makedirs(cache_path)

                # dir_path = os.path.dirname(os.path.realpath(__file__))
                pkl_path = f'{cache_path}/{func_name}.pkl'
                if os.path.exists(pkl_path) and not flushing:
                    with open(pkl_path, 'rb') as file:
                        df_value = pd.compat.pickle_compat.load(file)
                        # df_value = pickle.load(file)

                    print(f'[INFO]: the value of function [{func_name}] has been obtained from {cache_path}!')
                    return df_value
            
            df_value = func(*args, **kwargs)
            
            if caching:
                with open(pkl_path, 'wb') as file:
                    pickle.dump(df_value, file)
                print(f'[INFO]: the value of function [{func_name}] has been cached under {cache_path}!')
            
            return df_value
        return wrapper

    if _func is None:
        return decorator
    else:
        return decorator(_func)



import pandas as pd
import io

def load_excel_as_df(ctx, excel_file, sheet_name=None):

    if sheet_name is None:
        sheet_name = 0
    
    last_update_sys = excel_file.properties['TimeLastModified']
    last_update = pd.Timestamp(last_update_sys, tz='Europe/London').tz_convert(tz='UTC') # assume that the sharepoint system time in as London time

    # print(f'{last_update = }')
    files_url = excel_file.properties['ServerRelativeUrl']

    # print(f'{last_update=}')
    # print(f'{files_url=}')

    # files_url = excel_file

    # ctx.web.get().execute_query()

    with io.BytesIO() as data:
        _myfile = (ctx.web.get_file_by_server_relative_path(files_url)
                    .download(data)
                    .execute_query() 
                    ) 
        # print(f"{type(_myfile) = }")

        data.seek(0)

        # cols_dtype = {'board_id': str, 
        #               'hvac_board_id': str}

        df_contact = pd.read_excel(data, 
                                #    dtype=cols_dtype,
                            sheet_name=sheet_name,)
        
    # _myfile = (ctx.web.get_file_by_server_relative_path(files_url)
    #                 .execute_query() 
    #                 ) 
        
    df_contact['last_update'] = last_update

    return df_contact




import matplotlib.patches as mpatches
import matplotlib.text as mpltext

from .ruler_anno_shink_cf import ruler_anno_shink_cf


cd = {"tomato": '#FF836A',"aquablue": '#6DC2B3',"peach": '#FED6D2',"darkgrey": '#9F9D9C',"potato": '#FEF8C8',
      "cyan": '#B6E4E1',"dimgray":'#696969',"Seafoam":'#87e0cf',"gainsboro":'#DCDCDC'}


def benchmarking_design(ax, consumption_mwh_cur, floor_sqm,
                        kwh_per_sqm_good = 10.33,
                        kwh_per_sqm_typical = 14.5,
                        fx = 0.2, # fancy box x-axis value
                        fy = 0.18, # fancy box y-axis value
                        fs = 0.6, # fancy box scale factor
                        fontsize = 13
                        ):


        mwh_good = floor_sqm * kwh_per_sqm_good / 1000
        mwh_typical = floor_sqm * kwh_per_sqm_typical / 1000
        
        mwh_btw_good_typical = mwh_typical - mwh_good
        
        key_values_list = [mwh_typical + mwh_btw_good_typical,
                           mwh_good - mwh_btw_good_typical,
                           consumption_mwh_cur + mwh_btw_good_typical/2,
                           consumption_mwh_cur - mwh_btw_good_typical/2]
        
        ticks_max = max(key_values_list)
        ticks_min = min(key_values_list)
        
        ticks_values_list = [ticks_min, mwh_good, mwh_typical, ticks_max]
        ticks_range = ticks_max - ticks_min
        
        ## ruler settings
        ru_x = 0  # the x-axis of benchmarking ruler
        ru_y = 0  # the x-axis of benchmarking ruler
        ru_w = 1  # the width of benchmarking ruler
        ru_h = 0.08  # the height of benchmarking ruler
        
        rp_edgecolor = 'k' # edgecolor of the ruler body parts
        rp_note_fontsize = fontsize - 5 # the fontsize of notes on ruler parts # please note that the other sizes except default 13 haven't been tested yet

        # ruler body - part one - good
        rp_x = ru_x         # the x-axis value of this part
        rp_y = ru_y - 0.07      # the y-axis value of this part
        rp_w = ru_w * (ticks_values_list[1]-ticks_values_list[0])/ticks_range   # the width value of this part
        rp_h = ru_h
        rp_c = cd["aquablue"]
        ruler_body_part_one = mpatches.Rectangle([fx+rp_x*fs, fy+rp_y*fs], 
                                             rp_w*fs, rp_h*fs, facecolor=rp_c, edgecolor=rp_edgecolor, lw=1)
        
      #   print('ruler_body_part_one rp_w*fs, rp_h*fs: ', rp_w*fs, rp_h*fs)
        rpa_x = rp_x + rp_w/2.0 # ruler body part annotation - x axis
        rpa_y = rp_y  # ruler body part annotation - y axis
        bar_width = rp_w*fs
        part = 'part_one'
        rpa_text = ruler_anno_shink_cf(bar_width, part)
        

        ax.annotate(rpa_text, (fx+fs*rpa_x, fx+fs*rpa_y), color='w', weight='bold', 
                    fontsize=rp_note_fontsize, ha='center', va='center')

        # ruler body - part two - normal
        rp_x = rp_x + rp_w     # the x-axis value of this part
        rp_y = ru_y - 0.07      # the y-axis value of this part
        rp_w = ru_w * (ticks_values_list[2]-ticks_values_list[1])/ticks_range   # the width value of this part
        rp_h = ru_h
        rp_c = cd["potato"]
        ruler_body_part_two = mpatches.Rectangle([fx+rp_x*fs, fy+rp_y*fs], 
                                             rp_w*fs, rp_h*fs, facecolor=rp_c, edgecolor=rp_edgecolor, lw=1)
        
      #   print('ruler_body_part_two rp_w*fs, rp_h*fs: ', rp_w*fs, rp_h*fs)
        rpa_x = rp_x + rp_w/2.0 # ruler body part annotation - x axis
        rpa_y = rp_y  # ruler body part annotation - y axis
        bar_width = rp_w*fs
        part = 'part_two'
        rpa_text = ruler_anno_shink_cf(bar_width, part)

        ax.annotate(rpa_text, (fx+fs*rpa_x, fx+fs*rpa_y), color='#9F9D9C', weight='bold', 
                    fontsize=rp_note_fontsize, ha='center', va='center')

        # ruler body - part three - poor
        rp_x = rp_x + rp_w     # the x-axis value of this part
        rp_y = ru_y - 0.07      # the y-axis value of this part
        rp_w = ru_w * (ticks_values_list[3]-ticks_values_list[2])/ticks_range   # the width value of this part
        rp_h = ru_h
        rp_c = cd["tomato"]
        ruler_body_part_thr = mpatches.Rectangle([fx+rp_x*fs, fy+rp_y*fs], 
                                             rp_w*fs, rp_h*fs, facecolor=rp_c, edgecolor=rp_edgecolor, lw=1)
        
      #   print('ruler_body_part_three rp_w*fs, rp_h*fs: ', rp_w*fs, rp_h*fs)
        rpa_x = rp_x + rp_w/2.0 # ruler body part annotation - x axis
        rpa_y = rp_y  # ruler body part annotation - y axis
        bar_width = rp_w*fs
        part = 'part_three'
        rpa_text = ruler_anno_shink_cf(bar_width, part)

        ax.annotate(rpa_text, (fx+fs*rpa_x, fx+fs*rpa_y), color='w', weight='bold', 
                    fontsize=rp_note_fontsize, ha='center', va='center')

        ax.add_artist(ruler_body_part_one)
        ax.add_artist(ruler_body_part_two)
        ax.add_artist(ruler_body_part_thr)


        # ruler ticks labels
        
        rt_fs = fontsize  # the fontsize of ruler ticks labels
        rt_c = "grey"  # the font color of ruler ticks labels
        rt_va = u'bottom' 
        rt_ha = u'center' 
        rt_fontweight = "light"
        
        ru_rt_y = -0.18 # ruler ticks y-axis in relation to the ruler y-axis
        
        tick = ticks_values_list[0]
        rt_x = ru_x + (tick-ticks_min)/ticks_range
        rt_y = ru_y + ru_rt_y
        tick_value = f"{tick:1.1f}"
        tick_label_one = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{tick_value}', 
                                  va=rt_va, color=rt_c, ha=rt_ha, fontweight=rt_fontweight, fontsize=rt_fs)

        tick = ticks_values_list[1]
        tick_pct = (tick-ticks_min)/ticks_range
        rt_x = ru_x + tick_pct
        rt_y = ru_y + ru_rt_y

        # conditionally remove the tick_value label when it's too close to the edges
        if (tick_pct < 0.2) or (tick_pct > 0.8):
            tick_value = ''
        else:
            tick_value = f"{tick:1.1f}"
        tick_label_two = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{tick_value}', 
                                  va=rt_va, color=rt_c, ha=rt_ha, fontweight=rt_fontweight, fontsize=rt_fs)

        tick = ticks_values_list[2]
        tick_pct = (tick-ticks_min)/ticks_range
        rt_x = ru_x + tick_pct
        rt_y = ru_y + ru_rt_y

        # conditionally remove the tick_value label when it's too close to the edges
        if (tick_pct < 0.2) or (tick_pct > 0.8):
            tick_value = ''
        else:
            tick_value = f"{tick:1.1f}"
        tick_label_thr = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{tick_value}', 
                                  va=rt_va, color=rt_c, ha=rt_ha, fontweight=rt_fontweight, fontsize=rt_fs)

        tick = ticks_values_list[3]
        rt_x = ru_x + (tick-ticks_min)/ticks_range
        rt_y = ru_y + ru_rt_y
        tick_value = f"{tick:1.1f}"
        tick_label_four = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{tick_value}', 
                                  va=rt_va, color=rt_c, ha=rt_ha, fontweight=rt_fontweight, fontsize=rt_fs)

        ax.add_artist(tick_label_one)
        ax.add_artist(tick_label_two)
        ax.add_artist(tick_label_thr)
        ax.add_artist(tick_label_four)

        # ruler indicator
        indicator_value = consumption_mwh_cur
        rih_x = ru_x + (indicator_value-ticks_min)/ticks_range # x-axis value of the ruler indicator head
        rih_y = ru_y + 0.084  # y-axis value of the ruler indicator head
        rih_r = 0.045

        indicator_head = mpatches.Circle((fx+(rih_x)*fs, fy+rih_y*fs), rih_r*fs, 
                                         fc="w", color = "w", ec="k", lw=1.3, zorder=10
                                         ) #path_effects=[path_effects.withSimplePatchShadow()]

        rib_x = ru_x + (indicator_value-ticks_min)/ticks_range # x-axis value of the ruler indicator head
        rib_y = ru_y - 0.09  # y-axis value of the ruler indicator head
        rib_w = 0.02
        rib_h = ru_h * 2

        indicator_body = mpatches.Rectangle([fx+(rib_x-rib_w/2)*fs, fy+rib_y*fs], rib_w*fs, rib_h*fs, 
                                            fc="w", ec="k", lw=1.3,  zorder=10) #path_effects=[path_effects.withSimplePatchShadow()]
        ax.add_artist(indicator_body)
        ax.add_artist(indicator_head)
        
        ## title

        tt_x = ru_x + ru_w/2
        tt_y = ru_y - 0.26
        #title = f"BBP Benchmarking (REEB)"
        #tt_c = '#9F9D9C'
        #title_text = mpltext.Text(x=fx+tt_x*fs, y=fy+tt_y*fs, text=f'{title}', 
                                  #va="center", ha="center", color=tt_c, fontweight="normal", fontsize=12) #fontfamily='serif'
        
        #ax.add_artist(title_text)


import matplotlib.pyplot as plt

from .set_x_label_prepare_xtick import set_x_label_prepare_xtick
from .setup_y_tick_lim_label_zero_line import setup_y_tick_lim_label_zero_line
from .set_x_tick_and_lim import set_x_tick_and_lim

from mdl.utils_ import get_pos_bot_range_unit_format
from mdl.utils_ import plot_stacked_barchart

def plot_shifted_period_barchart(df_grouped_working_hours_period_unstacked, 
                                          fs=None,
                                          fontsize=None):

    if fs is None:
        fs = (3.51, 3.07)

    plt.style.use('seaborn-v0_8-white')
    fig, ax = plt.subplots(1, 1, figsize=fs)        

    df_grouped_working_hours_period_reset_with_sign = df_grouped_working_hours_period_unstacked.reset_index(drop=True)
    freqstr = df_grouped_working_hours_period_unstacked.index.freqstr

    df_grouped_working_hours_sorted_reset_pos, \
        df_grouped_working_hours_sorted_reset_bot, \
            ylim, \
                display_unit, \
                    formatter = get_pos_bot_range_unit_format(df_grouped_working_hours_period_reset_with_sign, 
                                                                        pad_pct_above_max=0.2)

    setup_y_tick_lim_label_zero_line(ax, ylim, display_unit, formatter, 
                                    freqstr=freqstr,
                                    fontsize=fontsize)

    x_ticks_labels = set_x_label_prepare_xtick(ax, df_grouped_working_hours_period_unstacked,
                                               fontsize=fontsize)

    plot_stacked_barchart(ax, df_grouped_working_hours_sorted_reset_pos, 
                        bottom=df_grouped_working_hours_sorted_reset_bot,
                        tick_range=ylim)

    set_x_tick_and_lim(ax, x_ticks_labels, df_grouped_working_hours_period_unstacked, 
                       fontsize=fontsize)

    fig.tight_layout(rect =(0.01, 0, 1, 1))
    subplots_adjust_left = None
    subplots_adjust_right = 0.68 # 0.92 # 

    fig.subplots_adjust(left=subplots_adjust_left,
                        right=subplots_adjust_right)


from typing import List, Union

def _searchConfColumnIdsMatch(self, columns, includeList:List[str], excludeList:List[str] = [])-> Union[str,None]:
    for c in columns:
        title:str = c["title"]
        id = c["id"]
        flag = True
        for item in includeList:
            if item.lower() not in title.lower():
                flag = False
                break
        if(not flag): continue
        flag = True
        for item in excludeList:
            if item.lower() in title.lower():
                flag = False
                break
        if(not flag): continue
        return id
    return None


from .items import items

def items_page(column_values_ids):
    items_page = f"""
            items_page(limit:500){{
                                cursor
                                {items(column_values_ids)}
                            }}
            """
    return items_page


import os

from . import databricks
from . import local

def get_credential_url():

    for local_credential_url in local.credential_urls:
        if os.path.exists(local_credential_url):
            return local_credential_url
        
    for databricks_credential_url in databricks.credential_urls:
        if os.path.exists(databricks_credential_url):
            return databricks_credential_url
        
    print("[INFO]: Cannot find credentials from local server or databricks!")
    return None




from .sync_with_monday import sync_with_monday
from .load_cf_from_db import load_cf_from_db

def fetch_monday_info(engine, 
                        monday_token,
                        table_name_cache, 
                        table_columns_cache,
                        monday_board_id,
                        columns_concerned=None, 
                        email_prefix=None,
                        sync_monday=None):

    if sync_monday is None:
        sync_monday = True
    
    # get contacts from db
    df_monday = load_cf_from_db(engine, table_name_cache, table_columns_cache)

    if sync_monday:
        df_monday = sync_with_monday(engine, 
                                    df_monday,
                                    monday_token, 
                                    table_name_cache, 
                                    monday_board_id,
                                    columns_concerned=columns_concerned, 
                                    email_prefix=email_prefix)
    return df_monday



from .get_sum_for_curr_and_prev import get_sum_for_curr_and_prev
from .get_subgroup_sum_and_rank import get_subgroup_sum_and_rank
from .get_rank_overall_protect import get_rank_overall_protect
from .get_entries_to_display import get_entries_to_display
from .get_change_pct_and_sort_and_trim_cols import get_change_pct_and_sort_and_trim_cols


def transform_asset_stats_for_adaptive_piechart(df_pivot_asset_group_by_period,
                                                      minimum_cnt_for_protection=None,
                                                      max_display_cnt=None,
                                                      sum_column = None, 
                                                      pct_column = None,
                                                      gene_column=None,
                                                      asset_group=None):

    df_pivot_asset_group_by_period_renamed = get_sum_for_curr_and_prev(df_pivot_asset_group_by_period)
    df_pivot_asset_group_by_period_renamed = get_subgroup_sum_and_rank(df_pivot_asset_group_by_period_renamed)
    df_pivot_asset_group_by_period_renamed_rdx = get_rank_overall_protect(df_pivot_asset_group_by_period_renamed, 
                                                                        minimum_cnt_for_protection=minimum_cnt_for_protection)
    df_pivot_asset_group_by_period_renamed_rdx_to_keep = get_entries_to_display(df_pivot_asset_group_by_period_renamed_rdx,
                                                                                    max_display_cnt=max_display_cnt,
                                                                                    gene_column=gene_column,
                                                                                    asset_group=asset_group)
    df_sort_idx_trim = get_change_pct_and_sort_and_trim_cols(df_pivot_asset_group_by_period_renamed_rdx_to_keep,
                                                sum_column = sum_column, 
                                                pct_column = pct_column,
                                                gene_column = gene_column,
                                                asset_group = asset_group)
    return df_sort_idx_trim



from .append_df_to_table import append_df_to_table
from mdl.utils_ import empty_table

def cache_df_to_db(engine, table_name, df):

    empty_table(engine, table_name, with_log=False)
    append_df_to_table(engine, table_name, df)

    return





from .load_energy_price import load_energy_price

def apply_price_from_db(engine, table_name_price, df_config_tz_sqm):

    df_price = load_energy_price(engine, table_name_price)
    df_config_tz_sqm = df_config_tz_sqm.rename(columns={'conv_mwh_price': 'conv_mwh_price_bak'})
    # df_config_tz_sqm_no_price = df_config_tz_sqm.drop(columns=['conv_mwh_price'])
    df_config_tz_price = df_config_tz_sqm.merge(df_price, on='building_id', how='left')
    df_config_tz_price['conv_mwh_price'] = df_config_tz_price['conv_mwh_price'].fillna(df_config_tz_price['conv_mwh_price_bak'])
    df_config_tz_price = df_config_tz_price.drop(columns=['conv_mwh_price_bak'])

    return df_config_tz_price



from .get_conn_by_id_json import get_conn_by_id_json

def get_conn_by_id(conn_id, 
                   account_key=None,
                   credential_url=None):
    
    # # fetching creds from local filesystem or datalake
    # cred_conn = get_conn_by_id_json(conn_id, 
    #                                 credential_url=credential_url)
    
    # if cred_conn != None:
    #     return cred_conn
    # else:
    #     # fetching creds from the airflow environment
    #     from .get_conn_by_id_airflow import get_conn_by_id_airflow
        
    #     cred_conn = get_conn_by_id_airflow(conn_id)

    # for testing
    cred_conn = None
    
    if cred_conn != None:
        return cred_conn
    else:
        # fetching creds from the synapse environment
        from .get_conn_by_id_synapse import get_conn_by_id_synapse

        if account_key is None:
            try:
                from notebookutils import mssparkutils
                account_key = mssparkutils.credentials.getConnectionStringOrCreds('AzureDataLakeStorageTest')
            except:
                raise Exception('Cannot get a valid account_key from mssparkutils, please provide one for getting credentials in the synapse environment')

        cred_conn = get_conn_by_id_synapse(conn_id, account_key)
        return cred_conn









from .load_building import load_building
from .load_org import load_org

from cache_return import cache_return

@cache_return
def attach_id_to_buildings(engine, df_config, table_org_name, table_building_name):

    df_org = load_org(engine, table_org_name)
    df_building = load_building(engine, table_building_name, caching=False)
    df_bld_org = df_building.merge(df_org, on='org_id', how='left').dropna()
    # df_config_bld_id = df_config.merge(df_bld_org, on=['building_name', 'org'], how='left')
    # df_config = df_config.drop(columns=['building_name', 'org'])
    df_config_bld_id = df_config.merge(df_bld_org, on='building_id', how='left')
    is_missing_bld_name = df_config_bld_id.building_id.isna().any()

    if is_missing_bld_name: # this part may not be very necessary after change the mergy on building_id
        df_config_bld_name_missing = df_config_bld_id.loc[df_config_bld_id['building_name'].isna()]

        missed_buildings = df_config_bld_name_missing['item_name'].to_list()
        raise Exception(f"The name(s) for building id(s) {missed_buildings} cannot be found!")

    return df_config_bld_id.drop(columns=['item_name', 'group_name'])




import pandas as pd

def patch_missing_columns_for_ooh_and_sign(df_pivot_working_hours_with_sign):

    columns_for_sign = [True, False]
    columns_for_ooh = [True, False]

    required_columns_for_ooh_and_sign = pd.MultiIndex.from_product([columns_for_sign, columns_for_ooh])

    # required_columns_for_ooh = [True, False]
    existing_columns_for_ooh_and_sign = df_pivot_working_hours_with_sign.columns

    for column in required_columns_for_ooh_and_sign:
        if column not in existing_columns_for_ooh_and_sign:
            df_pivot_working_hours_with_sign[column] = 0

    return df_pivot_working_hours_with_sign


import pandas as pd
import matplotlib
matplotlib.use('Agg') # turn off the display graphs 

from .preparing_source_data import preparing_source_data
from .generate_components import generate_components
from .generate_report import generate_report

def prepare_report(cf, site_obj, 
                   debug=False,
                   report_file_folder=None, 
                   directory_to_savefile=None):

    df_meta_with_value, df_meta_with_value_building = preparing_source_data(cf.postgresdb, 
                                                                            site_obj,
                                                                            debug=debug, 
                                                                            caching=False)

    # currently site_name is used for building_occupancy
    components = generate_components(cf.postgresdb, df_meta_with_value, df_meta_with_value_building, site_obj,
                                     fontsize=13,
                                     directory_to_savefile=directory_to_savefile)
    

    current_period = df_meta_with_value.period.max() # + pd.offsets.Week() 

    try: # handle the case where insights statement is not provided
        report_file_dict = generate_report(site_obj.site_name, current_period, 
                                           statements_list=site_obj.insight_statements, 
                                           organisation=site_obj.org_name,
                                           components=components,
                                           files_folder=directory_to_savefile,
                                           figures_folder=directory_to_savefile,
                                           report_file_folder=report_file_folder)
    except:
        report_file_dict = generate_report(site_obj.site_name, current_period, 
                                           organisation=site_obj.org_name,
                                           components=components,
                                           files_folder=directory_to_savefile,
                                           figures_folder=directory_to_savefile,
                                           report_file_folder=report_file_folder)

    return report_file_dict, current_period



import pandas as pd

def load_energy_price(engine, table_name):

    with engine.connect() as conn:

        query = f"""select space_id as building_id, price*1000 as conv_mwh_price, start as start_time from {table_name};""" # meta.space_annotations
        df_price_all = pd.read_sql_query(query, con=conn) 
    
    df_price_latest = df_price_all.loc[df_price_all.groupby('building_id').start_time.idxmax()].drop(columns='start_time')
        
    return df_price_latest

# def load_energy_price():
